from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from pydantic import ValidationError
from rich.live import Live

from chemex.configuration.experiment import ExperimentConfig, ExperimentNameConfig
from chemex.containers.experiment import Experiment
from chemex.containers.experiments import Experiments
from chemex.containers.profile import Profile
from chemex.experiments.factories import Creators, factories
from chemex.messages import (
    console,
    get_reading_exp_text,
    print_experiment_name_error,
    print_file_not_found_error,
    print_loading_experiments,
    print_pydantic_parsing_error,
)
from chemex.parameters import database
from chemex.parameters.factory import create_parameters
from chemex.toml import read_toml

if TYPE_CHECKING:
    from collections.abc import MutableMapping
    from pathlib import Path
    from typing import Any

    from chemex.configuration.methods import Selection
    from chemex.containers.dataset import Dataset

    ConfigType = ExperimentConfig[Any, Any]


def _get_experiment_name(config: MutableMapping[str, Any], filename: Path):
    try:
        return ExperimentNameConfig.model_validate(config).experiment.name
    except ValidationError:
        print_experiment_name_error(filename)
        sys.exit()


def _apply_selection(dataset: Dataset, selection: Selection) -> Dataset:
    if (include := selection.include) is not None:
        return [
            (spin_system, data)
            for spin_system, data in dataset
            if spin_system.part_of(include)
        ]

    if (exclude := selection.exclude) is not None:
        return [
            (spin_system, data)
            for spin_system, data in dataset
            if not spin_system.part_of(exclude)
        ]

    return dataset


def _create_dataset(
    filename: Path,
    live: Live,
    factory: Creators,
    config: ConfigType,
) -> Dataset:
    try:
        dataset = factory.create_dataset(filename.parent, config)
    except FileNotFoundError as e:
        live.stop()
        print_file_not_found_error(e)
        sys.exit(1)
    return dataset


def _create_config(
    filename: Path,
    live: Live,
    factory: Creators,
    config_dict: MutableMapping[str, Any],
) -> ConfigType:
    try:
        config = factory.create_config(config_dict)
    except ValidationError as e:
        live.stop()
        print_pydantic_parsing_error(filename, e)
        sys.exit()
    return config


def build_experiment(filename: Path, selection: Selection) -> Experiment:
    config_dict: MutableMapping[str, Any] = read_toml(filename)

    experiment_name = _get_experiment_name(config_dict, filename)

    with Live(get_reading_exp_text(filename, experiment_name), console=console) as live:
        factory = factories.get(experiment_name)

        config = _create_config(filename, live, factory, config_dict)

        printer = factory.create_printer()
        plotter = factory.create_plotter(filename, config)

        dataset = _create_dataset(filename, live, factory, config)
        dataset = _apply_selection(dataset, selection)

        profiles: list[Profile] = []

        for spin_system, data in dataset:
            spectrometer = factory.create_spectrometer(config, spin_system)
            sequence = factory.create_sequence(config)
            filterer = factory.create_filterer(config, spectrometer)
            name_map = create_parameters(config, spectrometer.liouvillian)
            profiles.append(
                Profile(
                    data,
                    spectrometer,
                    sequence,
                    name_map,
                    printer,
                    filterer,
                    config.data.scaled,
                ),
            )

        live.update(get_reading_exp_text(filename, experiment_name, len(profiles)))

    experiment = Experiment(filename, experiment_name, profiles, printer, plotter)
    experiment.estimate_noise(config.data.error)

    return experiment


def build_experiments(filenames: list[Path] | None, selection: Selection):
    if not filenames:
        return Experiments()

    print_loading_experiments()

    experiments = Experiments()

    for filename in filenames:
        experiment = build_experiment(filename, selection)
        experiments.add(experiment)

    database.sort_parameters()

    return experiments



from __future__ import annotations

from contextlib import ExitStack
from typing import TYPE_CHECKING, Any

import numpy as np
from matplotlib.backends.backend_pdf import PdfPages

from chemex.containers.data import Data
from chemex.messages import print_plot_filename
from chemex.plotters.plot import get_grid, plot_profile
from chemex.printers.plot import PlotPrinter, data_plot_printers

if TYPE_CHECKING:
    from pathlib import Path

    from chemex.containers.profile import Profile


def plot_relaxation(file_pdf: PdfPages, name: str, data_exp: Data, data_calc: Data):
    fig = plot_profile(name, data_exp, data_calc)
    ax2 = fig.axes[1]
    ax2.set_xlabel(r"Time (s)")
    ax2.set_ylabel(r"Intensity")
    ax2.set_xlim(0.0, max(data_exp.metadata) + min(data_calc.metadata))
    file_pdf.savefig(fig)


def create_plot_data_exp(profile: Profile) -> Data:
    data = profile.data

    times = data.metadata

    intensity0 = data.exp[np.argmax(abs(data.exp))]
    intensities = data.exp / intensity0
    intensities_calc = data.calc / intensity0

    errors = data.err / abs(intensity0)
    errorbars = np.array([-errors, errors]).transpose()

    data_exp = Data(exp=intensities, err=errorbars, metadata=times)
    data_exp.calc = intensities_calc
    data_exp.mask = data.mask
    data_exp.sort()

    return data_exp


def create_plot_data_calc(profile: Profile) -> Data:
    spectrometer = profile.spectrometer
    data = profile.data

    times = get_grid(data.metadata, 100, 0.02)

    filler = np.zeros_like(times)
    data_fit = Data(exp=filler, err=filler, metadata=times)

    intensity0 = data.exp[np.argmax(abs(data.exp))]
    scale = profile.data.scale / intensity0
    data_fit.calc = scale * profile.pulse_sequence.calculate(spectrometer, data_fit)

    return data_fit


class RelaxationPlotter:
    def __init__(self, filename: Path, **_extra: Any) -> None:
        self.filename = filename
        self.printer: PlotPrinter = data_plot_printers["relaxation"]

    def plot(self, path: Path, profiles: list[Profile]) -> None:
        basename = path / self.filename.name
        name_pdf = basename.with_suffix(".pdf")
        name_exp = basename.with_suffix(".exp")
        name_fit = basename.with_suffix(".fit")

        print_plot_filename(name_pdf)

        with ExitStack() as stack:
            file_pdf = stack.enter_context(PdfPages(str(name_pdf)))
            file_calc = stack.enter_context(name_fit.open("w"))
            file_exp = stack.enter_context(name_exp.open("w"))
            for profile in sorted(profiles):
                data_exp = create_plot_data_exp(profile)
                data_calc = create_plot_data_calc(profile)
                plot_relaxation(file_pdf, str(profile.spin_system), data_exp, data_calc)
                file_exp.write(
                    self.printer.print_exp(str(profile.spin_system), data_exp)
                )
                file_calc.write(
                    self.printer.print_calc(str(profile.spin_system), data_calc)
                )

    def plot_simulation(self, path: Path, profiles: list[Profile]) -> None:
        basename = path / self.filename.name
        name_pdf = basename.with_suffix(".pdf")
        name_sim = basename.with_suffix(".sim")

        print_plot_filename(name_pdf, extra=False)

        with ExitStack() as stack:
            file_pdf = stack.enter_context(PdfPages(str(name_pdf)))
            file_sim = stack.enter_context(name_sim.open("w"))
            for profile in sorted(profiles):
                data_exp = create_plot_data_exp(profile)
                data_calc = create_plot_data_calc(profile)
                plot_relaxation(file_pdf, str(profile.spin_system), data_exp, data_calc)
                file_sim.write(
                    self.printer.print_calc(str(profile.spin_system), data_calc)
                )



from __future__ import annotations

from functools import cache
from typing import TYPE_CHECKING

from chemex.models.factory import model_factory
from chemex.models.model import model
from chemex.parameters import database
from chemex.parameters.setting import LocalSettings, Parameters, ParamSetting
from chemex.parameters.spins import build_spin_param_settings

if TYPE_CHECKING:
    from typing import Any

    from chemex.configuration.conditions import Conditions
    from chemex.configuration.experiment import ExperimentConfig
    from chemex.nmr.basis import Basis
    from chemex.nmr.liouvillian import LiouvillianIS
    from chemex.parameters.spin_system import SpinSystem

    ConfigType = ExperimentConfig[Any, Any]


@cache
def _build_settings(
    basis: Basis,
    conditions: Conditions,
) -> tuple[LocalSettings, LocalSettings]:
    settings_spins, settings_spins_mf = build_spin_param_settings(basis, conditions)
    settings_kinetics = model_factory.create(model.name, conditions)
    settings = settings_kinetics | settings_spins
    settings_mf = settings_kinetics | settings_spins_mf
    return settings, settings_mf


def _set_to_fit(
    parameters: Parameters,
    name_map: dict[str, str],
    fitted: list[str],
) -> None:
    pool = set(name_map)
    for fitted_name in fitted:
        selection = {name for name in pool if name.startswith(fitted_name)}
        for name in selection:
            param_id = name_map[name]
            parameters[param_id].vary = True
            parameters[param_id].expr = ""
        pool -= selection


def _build_parameters(
    settings: LocalSettings,
    spin_system: SpinSystem,
    conditions: Conditions,
) -> tuple[dict[str, str], Parameters]:
    param_names = {
        local_name: setting.name_setting.get_param_name(spin_system, conditions)
        for local_name, setting in settings.items()
    }

    name_map = {
        local_name: param_name.id_ for local_name, param_name in param_names.items()
    }

    parameters: Parameters = {}
    for local_name, setting in settings.items():
        param_id = name_map[local_name]
        param_name = param_names[local_name]
        expression = setting.expr.format_map(name_map)
        parameters[param_id] = ParamSetting(
            param_name=param_name,
            value=setting.value,
            min=setting.min,
            max=setting.max,
            vary=setting.vary,
            expr=expression,
        )

    return name_map, parameters


def create_parameters(config: ConfigType, liouvillian: LiouvillianIS) -> dict[str, str]:
    # A copy is done because the output of '_build_settings' is cached
    settings, settings_mf = _build_settings(liouvillian.basis, config.conditions)

    name_map, parameters = _build_parameters(
        settings,
        liouvillian.spin_system,
        config.conditions,
    )

    name_map_mf, parameters_mf = _build_parameters(
        settings_mf,
        liouvillian.spin_system,
        config.conditions,
    )

    _set_to_fit(parameters, name_map, config.to_be_fitted.rates)
    _set_to_fit(parameters_mf, name_map_mf, config.to_be_fitted.model_free)

    database.add_parameters(parameters)
    database.add_parameters_mf(parameters_mf)

    selection = set(name_map) & liouvillian.basis.required_names

    return {local_name: name_map[local_name] for local_name in selection}



from __future__ import annotations

from functools import total_ordering
from re import search
from typing import TYPE_CHECKING

from .constants import AAA_TO_A

if TYPE_CHECKING:
    from collections.abc import Hashable


NO_NUMBER = -100000000


@total_ordering
class Group:
    """Represents a molecular group in a spin system.

    This class encapsulates the information about an amino acid residue or nucleic
    acid base, including its symbol, sequence number, and suffix for identifying
    different states. It provides methods for parsing group names and comparing group
    objects.

    Attributes:
        symbol (str): The symbol of the group.
        number (int): The number associated with the group, if any.
        suffix (str): Any suffix associated with the group name.
        search_keys (set[Hashable]): A set of keys used for searching or matching the
                                     group.
    """

    def __init__(self, name: str) -> None:
        """Initializes a Group object with a given name.

        The constructor parses the input name into symbol, number, and suffix
        components.

        Args:
            name (str): The name of the group to be parsed.
        """
        self.symbol, self.number, self.suffix = self.parse_group(name.strip().upper())
        self.symbol = AAA_TO_A.get(self.symbol, self.symbol)
        self.search_keys: set[Hashable] = {self} if self else set()

    def parse_group(self, name: str) -> tuple[str, int, str]:
        """Parses the group name into symbol, number, and suffix.

        Args:
            name (str): The name of the group to be parsed.

        Returns:
            tuple[str, int, str]: A tuple containing the symbol, number, and suffix.
        """
        if found := search("[0-9]+", name.strip().upper()):
            return name[: found.start()], int(found.group()), name[found.end() :]
        return name, NO_NUMBER, ""

    @property
    def name(self) -> str:
        """Represents the full name of the group.

        Returns:
            str: The full name of the group, combining symbol, number, and suffix.
        """
        number = "" if self.number == NO_NUMBER else self.number
        return f"{self.symbol}{number}{self.suffix}"

    def match(self, other: Group) -> bool:
        """Determines if this group matches another group.

        Args:
            other (Group): Another group to compare with.

        Returns:
            bool: True if the groups match, False otherwise.
        """
        symbol = other.symbol == self.symbol or not self.symbol
        number = self.number in (other.number, NO_NUMBER)
        suffix = other.suffix == self.suffix or not self.suffix
        return number and symbol and suffix

    def __lt__(self, other: object) -> bool:
        """Less than comparison for ordering of Group objects.

        Args:
            other (object): Another object to compare with.

        Returns:
            bool: True if this group is considered less than the other, False otherwise.
        """
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.number < other.number

    def __eq__(self, other: object) -> bool:
        """Equality comparison for Group objects.

        Args:
            other (object): Another object to compare with.

        Returns:
            bool: True if the groups are equal, False otherwise.
        """
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.name == other.name

    def __hash__(self) -> int:
        """Generates a hash value for a Group object.

        Returns:
            int: The hash value of the group.
        """
        return hash(self.name)

    def __bool__(self) -> bool:
        """Boolean representation of the Group object.

        Returns:
            bool: True if the group has a valid name, False otherwise.
        """
        return bool(self.name)

    def __str__(self) -> str:
        """String representation of the Group object.

        Returns:
            str: The full name of the group.
        """
        return self.name



from __future__ import annotations

import sys
from dataclasses import dataclass
from typing import TYPE_CHECKING, Annotated, Literal

from pydantic import BeforeValidator, ConfigDict, Field, ValidationError
from pydantic.types import PositiveInt

from chemex.configuration.base import BaseModelLowerCase
from chemex.messages import print_method_error
from chemex.parameters.spin_system import PydanticSpinSystem
from chemex.toml import read_toml

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path


# Type definitions
AllType = Annotated[Literal["*", "all"], BeforeValidator(str.lower)]
SelectionType = list[PydanticSpinSystem] | AllType | None


class Statistics(BaseModelLowerCase):
    mc: PositiveInt | None = None
    bs: PositiveInt | None = None
    bsn: PositiveInt | None = None


@dataclass
class Selection:
    include: SelectionType
    exclude: SelectionType


class Method(BaseModelLowerCase):
    model_config = ConfigDict(str_to_lower=True, extra="forbid")
    fitmethod: str = "leastsq"
    include: SelectionType = None
    exclude: SelectionType = None
    fit: list[str] = Field(default_factory=list)
    fix: list[str] = Field(default_factory=list)
    constraints: list[str] = Field(default_factory=list)
    grid: list[str] = Field(default_factory=list)
    statistics: Statistics | None = None

    @property
    def selection(self) -> Selection:
        return Selection(include=self.include, exclude=self.exclude)


Methods = dict[str, Method]


def read_methods(filenames: Iterable[Path]) -> Methods:
    methods: Methods = {}

    for filename in filenames:
        methods_dict = read_toml(filename)
        for section, settings in methods_dict.items():
            try:
                method = Method(**settings)
            except ValidationError as error:
                options = {option for err in error.errors() for option in err["loc"]}
                print_method_error(filename, section, options)
                sys.exit()
            methods[section] = method
    return methods



import sys
from argparse import Namespace
from collections.abc import Callable
from typing import Any

from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.backend_bases import Event
from matplotlib.figure import Figure
from matplotlib.widgets import Button, Slider

from chemex.configuration.methods import Selection
from chemex.containers.experiment import Experiment
from chemex.containers.experiments import Experiments
from chemex.experiments.builder import build_experiments
from chemex.experiments.loader import register_experiments
from chemex.models import model
from chemex.tools.pick_cest.buttons import Buttons

FigureSizeType = tuple[float, float]
AdjustParamsType = tuple[float, float, float, float]
PositionType = tuple[float, float, float, float]

FIGURE_SIZE = (12.0, 5.0)
ADJUST_PARAMS = (0.07, 0.1, 0.8, 0.9)
POSITION_SCALE = (0.855, 0.425, 0.1, 0.02)
POSITION_PREVIOUS = (0.825, 0.3, 0.075, 0.075)
POSITION_NEXT = (0.9, 0.3, 0.075, 0.075)
POSITION_SWAP = (0.825, 0.2, 0.15, 0.075)
POSITION_CLEAR = (0.825, 0.1, 0.15, 0.075)


def create_figure(
    size: FigureSizeType = FIGURE_SIZE,
    adjust_params: AdjustParamsType = ADJUST_PARAMS,
) -> tuple[Figure, Axes]:
    """Creates and returns a matplotlib figure and axes."""
    fig, axes = plt.subplots()
    fig.set_size_inches(*size)
    fig.subplots_adjust(*adjust_params)
    return fig, axes


def is_cest_experiment(experiment: Experiment) -> bool:
    """Checks if an experiment is a CEST experiment."""
    return experiment.name.startswith(("cest", "dcest", "coscest"))


def initialize_cest(args: Namespace) -> Experiments:
    """Initializes CEST experiments."""
    register_experiments()
    model.set_model("2st")
    return build_experiments(args.experiments, Selection(include=None, exclude=None))


def create_slider(
    fig: Figure,
    position: PositionType,
    label: str,
    val_range: tuple[float, float],
    initial_value: float,
    on_change: Callable[[float], None],
) -> Slider:
    """Creates and returns a matplotlib slider."""
    ax = fig.add_axes(position)
    slider = Slider(ax, label, *val_range, valfmt="%3.1f", valinit=initial_value)
    slider.on_changed(on_change)
    return slider


def create_button(
    fig: Figure,
    position: PositionType,
    label: str,
    on_click: Callable[[Event], Any],
) -> Button:
    """Creates and returns a matplotlib button."""
    ax = fig.add_axes(position)
    button = Button(ax, label)
    button.on_clicked(on_click)
    return button


def pick_cest(args: Namespace) -> None:
    """Pick peak positions in CEST profiles."""
    experiments = initialize_cest(args)
    sw = None

    for experiment in experiments:
        if not is_cest_experiment(experiment):
            sys.exit(
                f"\nError: '{experiment.name}' experiment not supported. "
                "The command 'chemex pick_cest' only works with CEST experiments.\n",
            )
        if experiment.name.startswith(("dcest", "coscest")):
            sw = 4.0

    fig, axes = create_figure()

    callback = Buttons(fig, axes, experiments, args.output, sw)

    if sw is not None:
        # For the Slider to remain responsive you must keep a reference to it.
        _slider = create_slider(
            fig,
            POSITION_SCALE,
            "Scale",
            (1.0, 10.0),
            sw,
            callback.set_sw,
        )

    # For the button to remain responsive you must keep a reference to it.
    _button1 = create_button(fig, POSITION_PREVIOUS, "Previous", callback.previous)
    _button2 = create_button(fig, POSITION_NEXT, "Next", callback.next)
    _button3 = create_button(fig, POSITION_SWAP, "Swap", callback.swap)
    _button4 = create_button(fig, POSITION_CLEAR, "Clear", callback.clear)

    fig.canvas.mpl_connect("button_press_event", callback.set_cs)

    plt.show()



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal

import numpy as np
from numpy.linalg import matrix_power

from chemex.configuration.data import RelaxationDataSettings
from chemex.configuration.experiment import CpmgSettings, ExperimentConfig, ToBeFitted
from chemex.containers.dataset import load_relaxation_dataset
from chemex.experiments.factories import Creators, factories
from chemex.filterers import PlanesFilterer
from chemex.nmr.basis import Basis
from chemex.nmr.liouvillian import LiouvillianIS
from chemex.nmr.spectrometer import Spectrometer
from chemex.plotters.cpmg import CpmgPlotter
from chemex.printers.data import CpmgPrinter

if TYPE_CHECKING:
    from chemex.containers.data import Data
    from chemex.parameters.spin_system import SpinSystem
    from chemex.typing import ArrayBool, ArrayFloat


EXPERIMENT_NAME = "cpmg_13c_ip"


class Cpmg13CIpSettings(CpmgSettings):
    name: Literal["cpmg_13c_ip"]
    time_t2: float
    carrier: float
    pw90: float
    time_equil: float = 0.0
    observed_state: Literal["a", "b", "c", "d"] = "a"

    @property
    def t_neg(self) -> float:
        return -2.0 * self.pw90 / np.pi

    @property
    def detection(self) -> str:
        return f"[iz_{self.observed_state}]"


class Cpmg13CIpConfig(ExperimentConfig[Cpmg13CIpSettings, RelaxationDataSettings]):
    @property
    def to_be_fitted(self) -> ToBeFitted:
        state = self.experiment.observed_state
        return ToBeFitted(rates=[f"r2_i_{state}"], model_free=[f"tauc_{state}"])


def build_spectrometer(
    config: Cpmg13CIpConfig,
    spin_system: SpinSystem,
) -> Spectrometer:
    settings = config.experiment
    conditions = config.conditions

    basis = Basis(type="ixyz", spin_system="ch")
    liouvillian = LiouvillianIS(spin_system, basis, conditions)
    spectrometer = Spectrometer(liouvillian)

    spectrometer.carrier_i = settings.carrier
    spectrometer.b1_i = 1 / (4.0 * settings.pw90)
    spectrometer.detection = settings.detection

    return spectrometer


@dataclass
class Cpmg13CIpSequence:
    settings: Cpmg13CIpSettings

    def _get_delays(self, ncycs: ArrayFloat) -> tuple[dict[float, float], list[float]]:
        ncycs_no_ref = ncycs[ncycs > 0]
        tau_cps = {
            ncyc: self.settings.time_t2 / (4.0 * ncyc) - self.settings.pw90
            for ncyc in ncycs_no_ref
        }
        # An ncyc of -1 corresponds to the experiment without the CPMG
        # refocusisng pulses
        tau_cps[-1.0] = 0.5 * self.settings.time_t2
        delays = [self.settings.t_neg, self.settings.time_equil, *tau_cps.values()]
        return tau_cps, delays

    def calculate(self, spectrometer: Spectrometer, data: Data) -> ArrayFloat:
        ncycs = data.metadata

        # Calculation of the spectrometers corresponding to all the delays
        tau_cps, all_delays = self._get_delays(ncycs)
        delays = dict(zip(all_delays, spectrometer.delays(all_delays), strict=True))
        d_neg = delays[self.settings.t_neg]
        d_eq = delays[self.settings.time_equil]
        d_cp = {ncyc: delays[delay] for ncyc, delay in tau_cps.items()}

        # Calculation of the spectrometers corresponding to all the pulses
        p90 = spectrometer.p90_i
        p180 = spectrometer.p180_i
        p180pmx = 0.5 * (p180[0] + p180[2])  # +/- phase cycling

        # Getting the starting magnetization
        start = spectrometer.get_equilibrium()

        # Calculating the instensities as a function of ncyc
        part1 = d_neg @ p90[0] @ start
        part2 = d_eq @ p90[0] @ d_neg
        intst = {
            0.0: spectrometer.detect(d_eq @ p90[0] @ p180pmx @ p90[0] @ start),
            -1.0: spectrometer.detect(part2 @ d_cp[-1] @ p180pmx @ d_cp[-1] @ part1),
        }
        for ncyc in set(ncycs) - {0.0, -1.0}:
            echo = d_cp[ncyc] @ p180[1] @ d_cp[ncyc]
            cpmg = matrix_power(echo, int(ncyc))
            intst[ncyc] = spectrometer.detect(part2 @ cpmg @ p180pmx @ cpmg @ part1)

        # Return profile
        return np.array([intst[ncyc] for ncyc in ncycs])

    @staticmethod
    def is_reference(metadata: ArrayFloat) -> ArrayBool:
        return metadata == 0


def register() -> None:
    creators = Creators(
        config_creator=Cpmg13CIpConfig,
        spectrometer_creator=build_spectrometer,
        sequence_creator=Cpmg13CIpSequence,
        dataset_creator=load_relaxation_dataset,
        filterer_creator=PlanesFilterer,
        printer_creator=CpmgPrinter,
        plotter_creator=CpmgPlotter,
    )
    factories.register(type=EXPERIMENT_NAME, creators=creators)



import contextlib
from pathlib import Path
from typing import Literal

import numpy as np
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.backend_bases import Event, LocationEvent
from matplotlib.figure import Figure
from matplotlib.widgets import Cursor

from chemex.containers.experiments import Experiments
from chemex.parameters.spin_system import SpinSystem
from chemex.parameters.spin_system.nucleus import Nucleus

from .curve import Curve

LSTYLE = {"a": "-", "b": ":"}
TEXT_Y = {"a": 0.8, "b": 0.75}
XLABELS = {
    Nucleus.N15: r"$^{15}$N (ppm)",
    Nucleus.C13: r"$^{13}$C (ppm)",
    Nucleus.H1: r"$^{1}$H (ppm)",
    Nucleus.X: r"Chemical shift (ppm)",
}


class Buttons:
    def __init__(
        self,
        figure: Figure,
        axis: Axes,
        experiments: Experiments,
        path: Path,
        sw: float | None,
    ) -> None:
        self.fig = figure
        self.axis = axis
        self.data: dict[SpinSystem, list[Curve]] = {}
        spin_systems: set[SpinSystem] = set()

        for experiment in experiments:
            for profile in experiment:
                spin_system = profile.spin_system
                spin_systems.add(spin_system)
                self.data.setdefault(spin_system, []).append(Curve(profile, sw))

        self.spin_systems = sorted(spin_systems)
        self.out = path

        self.spin_system = SpinSystem("")
        self.curves: list[Curve] = []
        self.cs_a: dict[SpinSystem, float | None] = {}
        self.cs_b: dict[SpinSystem, float | None] = {}
        self.artists: list[Artist] = []
        self.sw = sw

        self.cursor = Cursor(self.axis, horizOn=False, useblit=True)

        self.index = -1
        self.next()

    def _clear_artists(self) -> None:
        while self.artists:
            self.artists.pop().remove()

    def _clear_axis(self) -> None:
        self._clear_artists()
        self.axis.clear()

    def _show_labels(self) -> None:
        self.axis.set_title(str(self.spin_system))
        self.axis.set_xlabel(XLABELS[self.spin_system.nuclei["i"]])
        self.axis.set_ylabel("$I/I_0$")

    def _plot_profiles(self) -> None:
        if not self.curves:
            return
        xranges = np.concatenate([curve.get_xrange(self.sw) for curve in self.curves])
        grid = np.linspace(min(xranges), max(xranges), 1000)

        for index, curve in enumerate(self.curves):
            color = f"C{index}"
            self.axis.plot(curve.x, curve.y, ".", color=color)
            self.axis.plot(grid, curve.spline(grid), "--", color=color, alpha=0.66)

        self.axis.invert_xaxis()
        self._show_labels()
        self.fig.canvas.draw_idle()

    def _get_click_position(self, event: Event) -> float | None:
        if not isinstance(event, LocationEvent):
            return None
        if event.inaxes != self.axis:
            return None
        return event.xdata

    def _add_line(self, position: float, state: Literal["a", "b"]):
        text_ = rf"$\varpi_{state}$ = {position:.3f} ppm"
        text = self.fig.text(0.82, TEXT_Y[state], text_)
        line = self.axis.axvline(
            x=position,
            linestyle=LSTYLE[state],
            color="0.74",
            linewidth=1.0,
        )
        self.artists.extend([line, text])

    def _add_text_dw(self, dw_ab: float) -> None:
        text_ = rf"$\Delta\varpi_{{ab}}$ = {dw_ab:.3f} ppm"
        text = self.fig.text(0.82, 0.7, text_)
        self.artists.append(text)

    def _save(self):
        self.out.mkdir(parents=True, exist_ok=True)

        fname1 = self.out / "cs_a.toml"
        fname2 = self.out / "dw_ab.toml"

        with contextlib.ExitStack() as stack:
            file1 = stack.enter_context(fname1.open("w"))
            file2 = stack.enter_context(fname2.open("w"))
            file1.write("[CS_A]\n")
            file2.write("[DW_AB]\n")

            for name in self.spin_systems:
                cs_a = self.cs_a.get(name)
                cs_b = self.cs_b.get(name)

                if cs_a is None:
                    continue

                file1.write(f"{str(name).upper():10s} = {cs_a:8.3f}\n")

                if cs_b is not None:
                    dw_ab = cs_b - cs_a
                    file2.write(f"{str(name).upper():10s} = {dw_ab:8.3f}\n")

    def set_cs(self, event: Event) -> None:
        """Set the chemical shift."""
        xdata = self._get_click_position(event)
        if xdata is None:
            return

        key = self.spin_system

        if self.cs_a.get(key) is None or self.cs_b.get(key) is not None:
            self.cs_a[key] = xdata
            self.cs_b[key] = None
        else:
            self.cs_b[key] = xdata

        self._plot_lines()

    def _plot_lines(self) -> None:
        key = self.spin_system
        cs_a = self.cs_a.get(key)
        cs_b = self.cs_b.get(key)

        self._clear_artists()

        if cs_a is not None:
            self._add_line(cs_a, "a")

        if cs_b is not None:
            self._add_line(cs_b, "b")

        if cs_a is not None and cs_b is not None:
            self._add_text_dw(cs_b - cs_a)

        self._save()

        self.fig.canvas.draw_idle()
        self.fig.canvas.flush_events()

    def _plot(self, event: Event | None = None) -> None:
        self._clear_axis()
        self._plot_lines()
        self._plot_profiles()
        self.fig.canvas.draw_idle()

    def _shift(self, step: int):
        self.index += step
        self.index %= len(self.spin_systems)
        self.spin_system = self.spin_systems[self.index]
        self.curves = self.data[self.spin_system]
        self._clear_axis()
        self._plot()

    def next(self, _event: Event | None = None) -> None:
        """Go to next residue."""
        self._shift(+1)

    def previous(self, _event: Event) -> None:
        """Go to previous residue."""
        self._shift(-1)

    def swap(self, event: Event) -> None:
        """Swap peak peak positions for major/minor states."""
        name = self.spin_system
        if self.cs_b[name] is not None:
            self.cs_a[name], self.cs_b[name] = self.cs_b[name], self.cs_a[name]
        self._plot_lines()

    def clear(self, event: Event) -> None:
        name = self.spin_system
        self.cs_a[name], self.cs_b[name] = None, None
        self._plot_lines()

    def set_sw(self, sw: float) -> None:
        with contextlib.suppress(ValueError):
            self.sw = sw

        self._clear_axis()
        self._plot()



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from pathlib import Path

    from chemex.containers.profile import Profile


class Plotter(Protocol):
    def plot(self, path: Path, profiles: list[Profile]) -> None:
        ...

    def plot_simulation(self, path: Path, profiles: list[Profile]) -> None:
        ...



"""Module for managing and manipulating collections of NMR experiments.

This module contains the Experiments class and functions for generating different
types of simulated experiment datasets for statistical analysis.
"""
from __future__ import annotations

from itertools import chain
from random import choices
from typing import TYPE_CHECKING, Literal, TypeVar

import numpy as np

from chemex.messages import print_selecting_profiles
from chemex.parameters import database
from chemex.parameters.spin_system import Group, SpinSystem

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from lmfit.parameter import Parameters

    from chemex.configuration.methods import Selection
    from chemex.containers.experiment import Experiment
    from chemex.typing import ArrayFloat


# Type definitions
SelectionType = list[SpinSystem] | Literal["*", "all"], None

Self = TypeVar("Self", bound="Experiments")


class Experiments:
    """A collection of NMR experiments.

    This class manages a set of NMR experiments, providing functionalities for
    adding experiments, calculating residuals, simulation preparation, writing
    data, plotting, and more.
    """

    def __init__(self) -> None:
        """Initialize an empty Experiments collection."""
        self._experiments: dict[Path, Experiment] = {}

    @property
    def groups(self) -> set[Group]:
        """Get the set of unique groups across all experiments.

        Returns:
            set[Group]: A set of unique groups.
        """
        return {group for experiment in self for group in experiment.groups}

    def add(self, experiment: Experiment) -> None:
        """Add an experiment to the collection.

        Args:
            experiment (Experiment): The experiment to be added.
        """
        if experiment.filename in self._experiments:
            msg = "Experiment already exists."
            raise ValueError(msg)
        self._experiments[experiment.filename] = experiment

    def residuals(self, params: Parameters) -> ArrayFloat:
        """Calculate the residuals for all experiments in the collection.

        Args:
            params (Parameters): Parameters for residual calculation.

        Returns:
            ArrayFloat: Residuals as a NumPy array.
        """
        return np.asarray(
            list(
                chain.from_iterable(
                    experiment.residuals(params) for experiment in self
                ),
            ),
        )

    def back_calculate(self) -> None:
        """Back calculate experiments' parameters from the database."""
        params_lf = database.build_lmfit_params(self.param_ids)
        self.residuals(params_lf)

    def prepare_for_simulation(self) -> None:
        """Prepare each experiment in the collection for simulation."""
        self.back_calculate()
        for experiment in self:
            experiment.prepare_for_simulation()

    def write(self, path: Path):
        """Write experiment data to a specified path.

        Args:
            path (Path): Directory path to write data.
        """
        path_dat = path / "Data"
        path_dat.mkdir(parents=True, exist_ok=True)
        for experiment in self:
            experiment.write(path_dat)

    def plot(self, path: Path):
        """Plot each experiment in the collection.

        Args:
            path (Path): Directory path for saving plots.
        """
        for experiment in self:
            experiment.plot(path)

    def plot_simulation(self, path: Path):
        """Plot simulations for each experiment.

        Args:
            path (Path): Directory path for saving simulation plots.
        """
        for experiment in self:
            experiment.plot_simulation(path)

    def select(self, selection: Selection) -> None:
        """Select experiments based on the given selection criteria.

        Args:
            selection (Selection): Criteria for selecting experiments.
        """
        if selection.include is None and selection.exclude is None:
            return
        for experiment in self:
            experiment.select(selection)
        print_selecting_profiles(len(self))

    @property
    def param_id_sets(self) -> list[set[str]]:
        """Get a list of parameter ID sets for all experiments.

        Returns:
            list[set[str]]: List of sets containing parameter IDs.
        """
        return list(
            chain.from_iterable(experiment.param_id_sets for experiment in self),
        )

    @property
    def param_ids(self) -> set[str]:
        """Get a set of all unique parameter IDs across experiments.

        Returns:
            set[str]: Set of unique parameter IDs.
        """
        result: set[str] = set()
        return result.union(*(self.param_id_sets))

    def filter(self) -> None:
        """Get a set of all unique parameter IDs across experiments.

        Returns:
            set[str]: Set of unique parameter IDs.
        """
        params = database.build_lmfit_params(self.param_ids)
        for experiment in self:
            experiment.filter(params)

    def get_relevant_subset(self: Self, param_ids: set[str]) -> Self:
        """Get a subset of experiments relevant to specified parameter IDs.

        Args:
            param_ids (set[str]): Parameter IDs to filter experiments.

        Returns:
            Self: Subset of Experiments relevant to given parameter IDs.
        """
        relevant_subset = type(self)()
        for experiment in self:
            if subset := experiment.get_relevant_subset(param_ids):
                relevant_subset.add(subset)
        return relevant_subset

    def __iter__(self) -> Iterator[Experiment]:
        """Iterate over the experiments in the collection.

        Returns:
            Iterator[Experiment]: An iterator over experiments.
        """
        return iter(self._experiments.values())

    def __len__(self) -> int:
        """Get the total number of data points in the collection.

        Returns:
            int: Total number of data points.
        """
        return sum(len(experiment) for experiment in self)

    def __bool__(self) -> bool:
        """Check if the collection contains any experiments.

        Returns:
            bool: True if collection is non-empty, False otherwise.
        """
        return bool(self._experiments)


def generate_monte_carlo_experiments(experiments: Experiments) -> Experiments:
    """Generate a new Experiments collection for Monte Carlo simulation.

    Args:
        experiments (Experiments): Original experiments.

    Returns:
        Experiments: Collection with Monte Carlo simulated data.
    """
    experiments_mc = Experiments()
    for experiment in experiments:
        experiments_mc.add(experiment.monte_carlo())
    return experiments_mc


def generate_bootstrap_experiments(experiments: Experiments) -> Experiments:
    """Generate a new Experiments collection for Bootstrap simulation.

    Args:
        experiments (Experiments): Original experiments.

    Returns:
        Experiments: Collection with Bootstrap simulated data.
    """
    experiments_mc = Experiments()
    for experiment in experiments:
        experiments_mc.add(experiment.bootstrap())
    return experiments_mc


def generate_bootstrap_ns_experiments(experiments: Experiments) -> Experiments:
    """Generate a new Experiments collection for Bootstrap NS simulation.

    Args:
        experiments (Experiments): Original experiments.

    Returns:
        Experiments: Collection with Bootstrap NS simulated data.
    """
    groups = experiments.groups
    groups_bs = choices(tuple(groups), k=len(groups))
    experiments_bs = Experiments()
    for experiment in experiments:
        experiments_bs.add(experiment.bootstrap_ns(groups_bs))
    return experiments_bs


def generate_exp_for_statistics(
    experiments: Experiments,
    statistic_name: str,
) -> Experiments:
    """Generate experiments for a specific statistical method.

    Args:
        experiments (Experiments): Original experiments.
        statistic_name (str): Name of the statistical method.

    Returns:
        Experiments: Collection with simulated data for statistics.
    """
    generators = {
        "mc": generate_monte_carlo_experiments,
        "bs": generate_bootstrap_experiments,
        "bsn": generate_bootstrap_ns_experiments,
    }
    return generators[statistic_name](experiments)



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from functools import cache
from itertools import permutations, product
from typing import TYPE_CHECKING, Literal

import numpy as np

from chemex.models.model import model

if TYPE_CHECKING:
    from chemex.typing import ArrayFloat

    DictArrayFloat = dict[str, ArrayFloat]

_BASES = {
    "ixy": ["ix", "iy"],
    "iz": ["iz"],
    "izsz": ["iz", "2izsz"],
    "iz_eq": ["ie", "iz"],
    "ixyz": ["ix", "iy", "iz"],
    "ixyz_eq": ["ie", "ix", "iy", "iz"],
    "ixysxy": ["2ixsx", "2ixsy", "2iysx", "2iysy"],
    "ixy_ixysxy": ["ix", "iy", "2ixsx", "2ixsy", "2iysx", "2iysy"],
    "ixyzsz": ["ix", "iy", "iz", "2ixsz", "2iysz", "2izsz"],
    "ixyzsz_diff": ["ix", "iy", "iz", "2ixsz", "2iysz", "2izsz"],
    "ixyzsz_eq": ["ie", "ix", "iy", "iz", "2ixsz", "2iysz", "2izsz"],
    "ixyzsxyz": [
        "ix",
        "iy",
        "iz",
        "sx",
        "sy",
        "sz",
        "2ixsz",
        "2iysz",
        "2izsx",
        "2izsy",
        "2ixsx",
        "2ixsy",
        "2iysx",
        "2iysy",
        "2izsz",
    ],
    "ixyzsxyz_eq": [
        "ie",
        "se",
        "ix",
        "iy",
        "iz",
        "sx",
        "sy",
        "sz",
        "2ixsz",
        "2iysz",
        "2izsx",
        "2izsy",
        "2ixsx",
        "2ixsy",
        "2iysx",
        "2iysy",
        "2izsz",
    ],
}
_TRANSITIONS = {
    "r2_i_{state}": (("ix", "ix", -1.0), ("iy", "iy", -1.0)),
    "r2_s_{state}": (("sx", "sx", -1.0), ("sy", "sy", -1.0)),
    "r1_i_{state}": (("iz", "iz", -1.0), ("iz", "ie", +1.0)),
    "r1_s_{state}": (("sz", "sz", -1.0), ("sz", "se", +1.0)),
    "r2a_i_{state}": (("2ixsz", "2ixsz", -1.0), ("2iysz", "2iysz", -1.0)),
    "r2a_s_{state}": (("2izsx", "2izsx", -1.0), ("2izsy", "2izsy", -1.0)),
    "r2mq_is_{state}": (
        ("2ixsx", "2ixsx", -1.0),
        ("2ixsy", "2ixsy", -1.0),
        ("2iysx", "2iysx", -1.0),
        ("2iysy", "2iysy", -1.0),
    ),
    "r1a_is_{state}": (("2izsz", "2izsz", -1.0),),
    "etaxy_i_{state}": (
        ("ix", "2ixsz", -1.0),
        ("iy", "2iysz", -1.0),
        ("2ixsz", "ix", -1.0),
        ("2iysz", "iy", -1.0),
    ),
    "etaxy_s_{state}": (
        ("sx", "2izsx", -1.0),
        ("sy", "2izsy", -1.0),
        ("2izsx", "sx", -1.0),
        ("2izsy", "sy", -1.0),
    ),
    "etaz_i_{state}": (
        ("iz", "2izsz", -1.0),
        ("2izsz", "iz", -1.0),
        ("2izsz", "ie", +1.0),
    ),
    "etaz_s_{state}": (
        ("sz", "2izsz", -1.0),
        ("2izsz", "sz", -1.0),
        ("2izsz", "se", +1.0),
    ),
    "sigma_is_{state}": (
        ("iz", "sz", -1.0),
        ("sz", "iz", -1.0),
        ("sz", "ie", +1.0),
        ("iz", "se", +1.0),
    ),
    "mu_is_{state}": (
        ("2ixsx", "2iysy", +1.0),
        ("2ixsy", "2iysx", -1.0),
        ("2iysx", "2ixsy", -1.0),
        ("2iysy", "2ixsx", +1.0),
    ),
    "rotz_i": (
        ("ix", "iy", -1.0),
        ("iy", "ix", +1.0),
        ("2ixsx", "2iysx", -1.0),
        ("2iysx", "2ixsx", +1.0),
        ("2ixsy", "2iysy", -1.0),
        ("2iysy", "2ixsy", +1.0),
        ("2ixsz", "2iysz", -1.0),
        ("2iysz", "2ixsz", +1.0),
    ),
    "rotz_s": (
        ("sx", "sy", -1.0),
        ("sy", "sx", +1.0),
        ("2ixsx", "2ixsy", -1.0),
        ("2ixsy", "2ixsx", +1.0),
        ("2iysx", "2iysy", -1.0),
        ("2iysy", "2iysx", +1.0),
        ("2izsx", "2izsy", -1.0),
        ("2izsy", "2izsx", +1.0),
    ),
    "cs_i_{state}": (
        ("ix", "iy", -1.0),
        ("iy", "ix", +1.0),
        ("2ixsx", "2iysx", -1.0),
        ("2iysx", "2ixsx", +1.0),
        ("2ixsy", "2iysy", -1.0),
        ("2iysy", "2ixsy", +1.0),
        ("2ixsz", "2iysz", -1.0),
        ("2iysz", "2ixsz", +1.0),
    ),
    "cs_s_{state}": (
        ("sx", "sy", -1.0),
        ("sy", "sx", +1.0),
        ("2ixsx", "2ixsy", -1.0),
        ("2ixsy", "2ixsx", +1.0),
        ("2iysx", "2iysy", -1.0),
        ("2iysy", "2iysx", +1.0),
        ("2izsx", "2izsy", -1.0),
        ("2izsy", "2izsx", +1.0),
    ),
    "carrier_i": (
        ("ix", "iy", +1.0),
        ("iy", "ix", -1.0),
        ("2ixsx", "2iysx", +1.0),
        ("2iysx", "2ixsx", -1.0),
        ("2ixsy", "2iysy", +1.0),
        ("2iysy", "2ixsy", -1.0),
        ("2ixsz", "2iysz", +1.0),
        ("2iysz", "2ixsz", -1.0),
    ),
    "carrier_s": (
        ("sx", "sy", +1.0),
        ("sy", "sx", -1.0),
        ("2ixsx", "2ixsy", +1.0),
        ("2ixsy", "2ixsx", -1.0),
        ("2iysx", "2iysy", +1.0),
        ("2iysy", "2iysx", -1.0),
        ("2izsx", "2izsy", +1.0),
        ("2izsy", "2izsx", -1.0),
    ),
    "offset_i": (
        ("ix", "iy", +2.0 * np.pi),
        ("iy", "ix", -2.0 * np.pi),
        ("2ixsx", "2iysx", +2.0 * np.pi),
        ("2iysx", "2ixsx", -2.0 * np.pi),
        ("2ixsy", "2iysy", +2.0 * np.pi),
        ("2iysy", "2ixsy", -2.0 * np.pi),
        ("2ixsz", "2iysz", +2.0 * np.pi),
        ("2iysz", "2ixsz", -2.0 * np.pi),
    ),
    "offset_s": (
        ("sx", "sy", +2.0 * np.pi),
        ("sy", "sx", -2.0 * np.pi),
        ("2ixsx", "2ixsy", +2.0 * np.pi),
        ("2ixsy", "2ixsx", -2.0 * np.pi),
        ("2iysx", "2iysy", +2.0 * np.pi),
        ("2iysy", "2iysx", -2.0 * np.pi),
        ("2izsx", "2izsy", +2.0 * np.pi),
        ("2izsy", "2izsx", -2.0 * np.pi),
    ),
    "jeff_i": (
        ("ix", "iy", -2.0 * np.pi),
        ("iy", "ix", +2.0 * np.pi),
        ("2ixsx", "2iysx", -2.0 * np.pi),
        ("2iysx", "2ixsx", +2.0 * np.pi),
        ("2ixsy", "2iysy", -2.0 * np.pi),
        ("2iysy", "2ixsy", +2.0 * np.pi),
        ("2ixsz", "2iysz", -2.0 * np.pi),
        ("2iysz", "2ixsz", +2.0 * np.pi),
    ),
    "j_is_{state}": (
        ("ix", "2iysz", -np.pi),
        ("2iysz", "ix", +np.pi),
        ("2ixsz", "iy", -np.pi),
        ("iy", "2ixsz", +np.pi),
        ("sx", "2izsy", -np.pi),
        ("2izsy", "sx", +np.pi),
        ("2izsx", "sy", -np.pi),
        ("sy", "2izsx", +np.pi),
    ),
    "d_{state}": (
        ("ix", "ix", -1.0),
        ("iy", "iy", -1.0),
        ("iz", "iz", -1.0),
        ("sx", "sx", -1.0),
        ("sy", "sy", -1.0),
        ("sz", "sz", -1.0),
        ("2ixsz", "2ixsz", -1.0),
        ("2iysz", "2iysz", -1.0),
        ("2izsx", "2izsx", -1.0),
        ("2izsy", "2izsy", -1.0),
        ("2ixsx", "2ixsx", -1.0),
        ("2ixsy", "2ixsy", -1.0),
        ("2iysx", "2iysx", -1.0),
        ("2iysy", "2iysy", -1.0),
        ("2izsz", "2izsz", -1.0),
    ),
    "b1x_i": (
        ("iy", "iz", -2.0 * np.pi),
        ("iz", "iy", +2.0 * np.pi),
        ("2iysx", "2izsx", -2.0 * np.pi),
        ("2izsx", "2iysx", +2.0 * np.pi),
        ("2iysy", "2izsy", -2.0 * np.pi),
        ("2izsy", "2iysy", +2.0 * np.pi),
        ("2iysz", "2izsz", -2.0 * np.pi),
        ("2izsz", "2iysz", +2.0 * np.pi),
    ),
    "b1y_i": (
        ("iz", "ix", -2.0 * np.pi),
        ("ix", "iz", +2.0 * np.pi),
        ("2izsx", "2ixsx", -2.0 * np.pi),
        ("2ixsx", "2izsx", +2.0 * np.pi),
        ("2izsy", "2ixsy", -2.0 * np.pi),
        ("2ixsy", "2izsy", +2.0 * np.pi),
        ("2izsz", "2ixsz", -2.0 * np.pi),
        ("2ixsz", "2izsz", +2.0 * np.pi),
    ),
    "b1x_s": (
        ("sy", "sz", -2.0 * np.pi),
        ("sz", "sy", +2.0 * np.pi),
        ("2ixsy", "2ixsz", -2.0 * np.pi),
        ("2ixsz", "2ixsy", +2.0 * np.pi),
        ("2iysy", "2iysz", -2.0 * np.pi),
        ("2iysz", "2iysy", +2.0 * np.pi),
        ("2izsy", "2izsz", -2.0 * np.pi),
        ("2izsz", "2izsy", +2.0 * np.pi),
    ),
    "b1y_s": (
        ("sz", "sx", -2.0 * np.pi),
        ("sx", "sz", +2.0 * np.pi),
        ("2ixsz", "2ixsx", -2.0 * np.pi),
        ("2ixsx", "2ixsz", +2.0 * np.pi),
        ("2iysz", "2iysx", -2.0 * np.pi),
        ("2iysx", "2iysz", +2.0 * np.pi),
        ("2izsz", "2izsx", -2.0 * np.pi),
        ("2izsx", "2izsz", +2.0 * np.pi),
    ),
}
_ATOMS = {
    atoms: {"i": atoms[0], "s": atoms[1]} for atoms in ("hn", "hc", "nh", "ch", "cn")
}


@cache
def _build_vectors(basis: Basis) -> DictArrayFloat:
    size = len(basis) * len(model.states)
    vectors: defaultdict[str, ArrayFloat] = defaultdict(lambda: np.zeros((size, 1)))
    for index, (state, name) in enumerate(product(model.states, basis.components)):
        vectors[f"{name}_{state}"][index] = 1.0
        vectors[name][index] = 1.0
    return dict(vectors)


def _get_indices(
    basis: Basis,
    transition_name: str,
    state: str,
) -> tuple[tuple[list[int], list[int]], list[float]]:
    rows: list[int] = []
    cols: list[int] = []
    vals: list[float] = []
    offset = model.states.index(state) * len(basis)
    for start, end, value in _TRANSITIONS[transition_name]:
        if {start, end}.issubset(basis.components):
            rows.append(basis.components.index(start) + offset)
            cols.append(basis.components.index(end) + offset)
            vals.append(value)
    return (rows, cols), vals


def _build_spin_matrices(basis: Basis) -> DictArrayFloat:
    size = len(basis) * len(model.states)
    matrices: DictArrayFloat = defaultdict(lambda: np.zeros((size, size)))
    for transition_name, state in product(_TRANSITIONS, model.states):
        if not basis.type.endswith("_diff") and transition_name.startswith("d_"):
            continue
        name = transition_name.format(state=state)
        indices, values = _get_indices(basis, transition_name, state)
        if values:
            matrices[name][indices] = values
    return matrices


def _build_exchange_matrices(basis: Basis) -> DictArrayFloat:
    matrices: DictArrayFloat = {}
    for (i1, s1), (i2, s2) in permutations(enumerate(model.states), r=2):
        name = f"k{s1}{s2}"
        matrix = np.zeros((len(model.states), len(model.states)))
        matrix[((i1, i2), i1)] = -1.0, 1.0
        matrices[name] = np.kron(matrix, np.eye(len(basis)))
    return matrices


@cache
def _build_matrices(basis: Basis) -> DictArrayFloat:
    return _build_exchange_matrices(basis) | _build_spin_matrices(basis)


@dataclass(frozen=True)
class Basis:
    type: str
    extension: Literal["", "dq", "tq"] = ""
    spin_system: str = ""

    @property
    def name(self) -> str:
        return f"{self.type}.{self.extension}.{self.spin_system}"

    @property
    def components(self) -> list[str]:
        return _BASES[self.type]

    @property
    def atoms(self) -> dict[str, str]:
        return {
            letter: atom
            for letter, atom in _ATOMS.get(self.spin_system, {}).items()
            if letter in self.type
        }

    @property
    def vectors(self) -> DictArrayFloat:
        return dict(_build_vectors(self))

    @property
    def matrices(self) -> DictArrayFloat:
        return _build_matrices(self)

    @property
    def required_names(self) -> set[str]:
        required_names = set(self.matrices)
        required_names |= {f"p{state}" for state in model.states}
        return required_names

    def __len__(self) -> int:
        return len(self.components)



"""A simple plugin loader."""
from __future__ import annotations

import importlib
from pkgutil import iter_modules
from typing import Protocol, cast

from chemex.experiments import catalog


class ExperimentModuleInterface(Protocol):
    """Represents an experiment module interface.
    An experiment module plugin has a single register function.
    """

    @staticmethod
    def register() -> None:
        """Register the necessary items related to the experiment."""
        ...


def import_module(name: str) -> ExperimentModuleInterface:
    """Imports a module given a name."""
    return cast(ExperimentModuleInterface, importlib.import_module(name))


def register_experiments() -> None:
    """Loads the plugins defined in the plugins list."""
    for module in iter_modules(catalog.__path__):
        module_name = f"{catalog.__name__}.{module.name}"
        experiment_module = import_module(module_name)
        experiment_module.register()



"""The ref module contains the code for calculating the Liouvillians.

Operator basis:

{ Eq,
Ix, Iy, Iz, Sx, Sy, Sz,
2IxSz, 2IySz, 2IzSx, 2IzSy,
2IxSx, 2IxSy, 2IySx, 2IySy,
2IzSz }

"""
from __future__ import annotations

import re
from itertools import product
from typing import TYPE_CHECKING

import numpy as np
from scipy import stats

from chemex.models.model import model
from chemex.nmr.constants import SIGNED_XI_RATIO, XI_RATIO, Distribution

if TYPE_CHECKING:
    from collections.abc import Iterable

    from chemex.configuration.conditions import Conditions
    from chemex.nmr.basis import Basis
    from chemex.parameters.spin_system import SpinSystem
    from chemex.typing import ArrayFloat


_RE_COMP = re.compile(r"\[(.+?)\]")

_Q_ORDER_I = {"sq": 1.0, "dq": 2.0, "tq": 3.0}


def _make_gaussian(
    value: float,
    scale: float,
    res: int,
) -> tuple[ArrayFloat, ArrayFloat]:
    if scale not in (0.0, np.inf) and res > 1:
        grid = np.linspace(-2.0, 2.0, res)
        dist = grid * scale + 1.0
    else:
        grid = np.array([0.0])
        dist = np.array([1.0])
    weights = stats.norm.pdf(grid)
    weights /= weights.sum()
    return dist * value, weights


class LiouvillianIS:
    def _scale_matrix(self, name: str, value: float):
        state_names = {name.format(state=state) for state in model.states}
        for state_name in state_names & set(self.matrices):
            self.matrices[state_name] = np.sign(self._matrices_ref[state_name]) * value

    def __init__(
        self,
        spin_system: SpinSystem,
        basis: Basis,
        conditions: Conditions,
    ) -> None:
        self.spin_system = spin_system.correct(basis)
        self.basis = basis
        self.h_frq = 0.0 if conditions.h_larmor_frq is None else conditions.h_larmor_frq
        self.par_values: dict[str, float] = {}
        self.size = len(model.states) * len(basis)
        self.vectors = basis.vectors.copy()
        self.matrices = basis.matrices.copy()
        self._matrices_ref = self.matrices.copy()
        self._detection: str = ""
        self._detect_vector: ArrayFloat = np.array([])
        self._q_order_i = _Q_ORDER_I.get(self.basis.extension, 1.0)
        scale = -2.0 * np.pi * self.h_frq
        self.ppm_i = scale * SIGNED_XI_RATIO.get(basis.atoms.get("i", ""), 1.0)
        self.ppm_s = scale * SIGNED_XI_RATIO.get(basis.atoms.get("s", ""), 1.0)
        self.carrier_i = 0.0
        self.carrier_s = 0.0
        self.offset_i = 0.0
        self.offset_s = 0.0
        self._l_base = np.array(0.0)
        self._b1_i_inh_scale = 0.0
        self._b1_i_inh_res = 11
        self.b1_i = 1e32
        self.b1_s = 1e32
        self.jeff_i = Distribution(np.array([0.0]), np.array([1.0]))
        self.gradient_dephasing = 0.0
        self._scale_matrix("j_is_{state}", self._q_order_i * np.pi)

    def _build_base_liouvillian(self) -> None:
        self._l_base = sum(
            (
                self.matrices[name] * self.par_values.get(name, 0.0)
                for name in self.matrices
            ),
            start=np.array(0.0),
        )

    def update(self, par_values: dict[str, float]) -> None:
        self.par_values = par_values
        self._build_base_liouvillian()

    @property
    def ppm_i(self) -> float:
        return self._ppm_i

    @ppm_i.setter
    def ppm_i(self, value: float):
        self._ppm_i = value
        self._scale_matrix("cs_i_{state}", self._q_order_i * value)
        self._scale_matrix("carrier_i", self._q_order_i * value)

    @property
    def ppm_s(self) -> float:
        return self._ppm_s

    @ppm_s.setter
    def ppm_s(self, value: float):
        self._ppm_s = value
        self._scale_matrix("cs_s_{state}", value)
        self._scale_matrix("carrier_s", value)

    @property
    def carrier_i(self) -> float:
        return float(self._carrier_i)

    @carrier_i.setter
    def carrier_i(self, value: float):
        self._carrier_i = np.array(value)
        self._l_carrier_i = self.matrices.get("carrier_i", np.array(0.0)) * value

    @property
    def carrier_s(self) -> float:
        return float(self._carrier_s)

    @carrier_s.setter
    def carrier_s(self, value: float):
        self._carrier_s = np.array(value)
        self._l_carrier_s = self.matrices.get("carrier_s", np.array(0.0)) * value

    @property
    def offset_i(self) -> float:
        return float(self._offset_i)

    @offset_i.setter
    def offset_i(self, value: float):
        self._offset_i = np.array(value)
        self._l_offset_i = (
            self.matrices.get("offset_i", np.array(0.0)) * value * np.sign(self.ppm_i)
        )

    @property
    def offset_s(self) -> float:
        return float(self._offset_s)

    @offset_s.setter
    def offset_s(self, value: float):
        self._offset_s = np.array(value)
        self._l_offset_s = (
            self.matrices.get("offset_s", 0.0) * value * np.sign(self.ppm_s)
        )

    @property
    def b1_i_inh_scale(self) -> float:
        return self._b1_i_inh_scale

    @b1_i_inh_scale.setter
    def b1_i_inh_scale(self, value: float):
        self._b1_i_inh_scale = value
        self.b1_i = self._b1_i

    @property
    def b1_i_inh_res(self) -> int:
        return self._b1_i_inh_res

    @b1_i_inh_res.setter
    def b1_i_inh_res(self, value: int):
        self._b1_i_inh_res = value
        self.b1_i = self._b1_i

    @property
    def b1_i(self) -> float:
        return self._b1_i

    @b1_i.setter
    def b1_i(self, value: float):
        self._b1_i = value
        self._b1_i_dist, self._b1_i_weights = _make_gaussian(
            self.b1_i,
            self.b1_i_inh_scale,
            self.b1_i_inh_res,
        )
        self._b1_i_dist = self._b1_i_dist.reshape((-1, 1, 1))
        self._b1_i_weights = self._b1_i_weights.reshape((-1, 1, 1))
        self.l_b1x_i = self.matrices.get("b1x_i", 0.0) * self._b1_i_dist
        self.l_b1y_i = self.matrices.get("b1y_i", 0.0) * self._b1_i_dist

    @property
    def b1_s(self) -> float:
        return self._b1_s

    @b1_s.setter
    def b1_s(self, value: float):
        self._b1_s = value
        self.l_b1x_s = self.matrices.get("b1x_s", 0.0) * value
        self.l_b1y_s = self.matrices.get("b1y_s", 0.0) * value

    @property
    def jeff_i(self) -> Distribution:
        return self._jeff_i

    @jeff_i.setter
    def jeff_i(self, distribution: Distribution):
        self._jeff_i = distribution
        values = distribution.values.reshape((-1, 1, 1, 1))
        self._l_jeff_i = self.matrices.get("jeff_i", np.array(0.0)) * values
        self._jeff_i_weights = distribution.weights.reshape((-1, 1, 1, 1))

    @property
    def gradient_dephasing(self) -> float:
        return self._gradient_dephasing

    @gradient_dephasing.setter
    def gradient_dephasing(self, value: float):
        self._gradient_dephasing = value
        self._scale_matrix("d_{state}", value * 1e-12)
        self._build_base_liouvillian()

    @property
    def l_free(self) -> ArrayFloat:
        return sum(
            (
                self._l_base,
                self._l_offset_i,
                self._l_offset_s,
                self._l_carrier_i,
                self._l_carrier_s,
                self._l_jeff_i,
            ),
            start=np.array(0.0),
        )

    @property
    def weights(self) -> ArrayFloat:
        return self._b1_i_weights * self._jeff_i_weights

    @property
    def detection(self):
        return self._detection

    @detection.setter
    def detection(self, value: str):
        self._detection = value
        expr = _RE_COMP.sub(r'self.vectors.get("\g<1>")', value)
        vector: ArrayFloat = eval(expr)
        self._detect_vector = vector.transpose()

    def detect(self, magnetization: ArrayFloat) -> float:
        shape = -1, *magnetization.shape[-2:]
        mag_weighted = self.weights * magnetization
        mag = mag_weighted.reshape(shape).sum(axis=0)
        detected = self._detect_vector @ mag
        if np.iscomplexobj(detected):
            detected = np.sign(detected.real) * np.abs(detected)
        return float(detected)

    def get_equilibrium(self) -> ArrayFloat:
        mag = np.zeros((self.size, 1))
        for state, (name, atom) in product(model.states, self.basis.atoms.items()):
            scale = self.par_values.get(f"p{state}", 0.0) * XI_RATIO.get(atom, 1.0)
            mag += self.vectors.get(f"{name}e_{state}", 0.0) * scale
            mag += self.vectors.get(f"{name}z_{state}", 0.0) * scale
        return mag

    def get_start_magnetization(
        self,
        terms: Iterable[str],
        atom: str = "h",
    ) -> ArrayFloat:
        ratio = XI_RATIO.get(atom, 1.0)
        mag = np.zeros((self.size, 1))
        for term, state, (comp, vector) in product(
            terms,
            model.states,
            self.vectors.items(),
        ):
            if comp.startswith(term) and comp.endswith(f"_{state}"):
                mag += self.par_values[f"p{state}"] * ratio * vector
        return mag

    def keep(self, magnetization: ArrayFloat, components: Iterable[str]) -> ArrayFloat:
        keep = sum(
            (self.vectors[name] for name in components),
            start=np.zeros((self.size, 1)),
        )
        keep[keep > 0] = 1.0
        return keep * magnetization

    def offsets_to_ppms(self, offsets: ArrayFloat) -> ArrayFloat:
        return self.carrier_i + 2.0 * np.pi * offsets / abs(self.ppm_i)

    def ppms_to_offsets(self, ppms: ArrayFloat | float) -> ArrayFloat | float:
        return (ppms - self.carrier_i) * abs(self.ppm_i) / (2.0 * np.pi)



from __future__ import annotations

import re
from dataclasses import dataclass, field
from re import Pattern
from typing import TYPE_CHECKING, Literal

import numpy as np

from chemex.parameters.name import ParamName

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions
    from chemex.configuration.parameters import DefaultSetting
    from chemex.parameters.spin_system import SpinSystem

_RE_IDS = re.compile(r"__[a-zA-Z_][a-zA-Z0-9_]*")
_RE_NAMES = re.compile(r"\{([a-zA-Z0-9_]+)\}")


@dataclass(frozen=True)
class NameSetting:
    name: str
    spin_system_part: Literal["i", "s", "is", "g", ""] = ""
    conditions_part: tuple[str, ...] = field(default_factory=tuple)

    def get_param_name(
        self,
        spin_system: SpinSystem,
        conditions: Conditions,
    ) -> ParamName:
        param_spin_system = spin_system.build_sub_spin_system(self.spin_system_part)
        param_conditions = conditions.select_conditions(self.conditions_part)
        return ParamName(self.name, param_spin_system, param_conditions)


class ExpressionSetting:
    def __init__(self, re_dependencies: Pattern[str]) -> None:
        self.re_dependencies = re_dependencies
        self.__expr: str = ""
        self.__dependencies: set[str] = set()

    @property
    def expr(self) -> str:
        return self.__expr

    @expr.setter
    def expr(self, expression: str) -> None:
        self.__expr = expression
        self.__dependencies = set(self.re_dependencies.findall(expression))

    @property
    def dependencies(self) -> set[str]:
        return self.__dependencies

    def __str__(self) -> str:
        return self.__expr


class ParamLocalSetting:
    def __init__(
        self,
        name_setting: NameSetting,
        value: float | None = None,
        min: float = -np.inf,
        max: float = np.inf,
        vary: bool = False,
        expr: str = "",
    ) -> None:
        self.__expr: ExpressionSetting = ExpressionSetting(_RE_NAMES)
        self.name_setting = name_setting
        self.value = value
        self.min = min
        self.max = max
        self.vary = vary
        self.expr = expr

    @property
    def expr(self) -> str:
        return str(self.__expr)

    @expr.setter
    def expr(self, value: str) -> None:
        self.__expr.expr = value

    @property
    def dependencies(self) -> set[str]:
        return self.__expr.dependencies


class ParamSetting:
    def __init__(
        self,
        param_name: ParamName,
        value: float | None = None,
        stderr: float | None = None,
        min: float = -np.inf,
        max: float = np.inf,
        vary: bool = False,
        expr: str = "",
        brute_step: float | None = None,
    ) -> None:
        self.__expr: ExpressionSetting = ExpressionSetting(_RE_IDS)
        self.param_name = param_name
        self.value = value
        self.stderr = stderr
        self.min = min
        self.max = max
        self.vary = vary
        self.expr = expr
        self.brute_step = brute_step

    @property
    def expr(self) -> str:
        return str(self.__expr)

    @expr.setter
    def expr(self, value: str) -> None:
        self.__expr.expr = value

    @property
    def dependencies(self) -> set[str]:
        return self.__expr.dependencies

    @property
    def id_(self) -> str:
        return self.param_name.id_

    @property
    def args(self) -> tuple[str, float | None, bool | None, float, float, str]:
        return self.id_, self.value, self.vary, self.min, self.max, self.expr

    def set(self, default_setting: DefaultSetting) -> None:
        self.value = default_setting.value
        if default_setting.min is not None:
            self.min = default_setting.min
        if default_setting.max is not None:
            self.max = default_setting.max
        if default_setting.min is not None:
            self.brute_step = default_setting.brute_step


LocalSettings = dict[str, ParamLocalSetting]

Parameters = dict[str, ParamSetting]



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING

import numpy as np
from scipy.constants import constants

from chemex.models.constraints import pop_2st
from chemex.models.factory import model_factory
from chemex.parameters.setting import NameSetting, ParamLocalSetting
from chemex.parameters.userfunctions import user_function_registry

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions

NAME = "2st_eyring"

PL = ("p_total", "l_total")
TPL = ("temperature", "p_total", "l_total")


@lru_cache(maxsize=100)
def calculate_kij_2st_eyring(
    dh_b: float,
    ds_b: float,
    dh_ab: float,
    ds_ab: float,
    temperature: float,
) -> dict[str, float]:
    kelvin = temperature + 273.15
    kbt_h = constants.k * kelvin / constants.h
    rt = constants.R * kelvin
    dh_a = ds_a = 0.0
    kab = kbt_h * np.exp(-(dh_ab - dh_a - kelvin * (ds_ab - ds_a)) / rt)
    kba = kbt_h * np.exp(-(dh_ab - dh_b - kelvin * (ds_ab - ds_b)) / rt)
    return {"kab": kab, "kba": kba}


def make_settings_2st_eyring(conditions: Conditions) -> dict[str, ParamLocalSetting]:
    celsius = conditions.temperature
    return {
        "dh_b": ParamLocalSetting(
            name_setting=NameSetting("dh_b", "", PL),
            value=8e3,
            vary=True,
        ),
        "ds_b": ParamLocalSetting(
            name_setting=NameSetting("ds_b", "", PL),
            value=0.0,
            vary=True,
        ),
        "dh_ab": ParamLocalSetting(
            name_setting=NameSetting("dh_ab", "", PL),
            value=6.5e4,
            vary=True,
        ),
        "ds_ab": ParamLocalSetting(
            name_setting=NameSetting("ds_ab", "", PL),
            value=0.0,
            vary=True,
        ),
        "kab": ParamLocalSetting(
            name_setting=NameSetting("kab", "", TPL),
            min=0.0,
            expr=f"kij_2st_eyring({{dh_b}},{{ds_b}},{{dh_ab}},{{ds_ab}},{celsius})['kab']",
        ),
        "kba": ParamLocalSetting(
            name_setting=NameSetting("kba", "", TPL),
            min=0.0,
            expr=f"kij_2st_eyring({{dh_b}},{{ds_b}},{{dh_ab}},{{ds_ab}},{celsius})['kba']",
        ),
        "pa": ParamLocalSetting(
            name_setting=NameSetting("pa", "", TPL),
            min=0.0,
            max=1.0,
            expr="pop_2st({kab},{kba})['pa']",
        ),
        "pb": ParamLocalSetting(
            name_setting=NameSetting("pb", "", TPL),
            min=0.0,
            max=1.0,
            expr="pop_2st({kab},{kba})['pb']",
        ),
    }


def register() -> None:
    model_factory.register(name=NAME, setting_maker=make_settings_2st_eyring)
    user_functions = {"kij_2st_eyring": calculate_kij_2st_eyring, "pop_2st": pop_2st}
    user_function_registry.register(name=NAME, user_functions=user_functions)



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from typing import TYPE_CHECKING

import lmfit
import numpy as np
from lmfit import Minimizer, Parameters

from chemex.messages import (
    print_calculation_stopped_error,
    print_chi2_table_footer,
    print_chi2_table_header,
    print_chi2_table_line,
    print_value_error,
)
from chemex.optimize.grouping import ParamTree, create_group_tree

if TYPE_CHECKING:
    from chemex.containers.experiments import Experiments
    from chemex.typing import ArrayFloat


@dataclass
class Reporter:
    last_chisqr: float = 1e32
    threshold: float = -1.0e-3

    def iter_cb(
        self,
        params: Parameters,
        iteration: int,
        residuals: ArrayFloat,
        *_args,
        **_kwargs,
    ) -> None:
        chisqr = (residuals**2).sum()
        change = (chisqr - self.last_chisqr) / self.last_chisqr

        if change > self.threshold or iteration < 0:
            return

        self.last_chisqr = chisqr

        ndata = len(residuals)
        nvarys = len(
            [param for param in params.values() if param.vary and not param.expr],
        )
        redchi = chisqr / max(1, ndata - nvarys)

        self.print_line(iteration, chisqr, redchi)

    def print_line(self, iteration: int, chisqr: float, redchi: float) -> None:
        print_chi2_table_line(iteration, chisqr, redchi)

    def print_header(self) -> None:
        print_chi2_table_header()

    def print_footer(self, iteration: int, chisqr: float, redchi: float) -> None:
        print_chi2_table_footer(iteration, chisqr, redchi)


def minimize(
    experiments: Experiments,
    params: Parameters,
    fitmethod: str,
) -> Parameters:
    kws = {
        "brute": {"keep": "all"},
        "basinhopping": {"niter_success": 10},
    }

    minimizer = Minimizer(experiments.residuals, params)
    minimizer.minimize(method=fitmethod, **(kws.get(fitmethod, {})))
    return deepcopy(minimizer.result.params)


def minimize_with_report(
    experiments: Experiments,
    params: Parameters,
    fitmethod: str,
) -> Parameters:
    kws = {
        "brute": {"keep": "all"},
        "basinhopping": {"niter_success": 10},
        "differential_evolution": {"polish": True, "disp": True, "init": "sobol"},
        "ampgo": {"disp": True},
    }

    reporter = Reporter()

    minimizer = Minimizer(experiments.residuals, params, iter_cb=reporter.iter_cb)

    reporter.print_header()

    try:
        result = minimizer.minimize(method=fitmethod, **(kws.get(fitmethod, {})))
        reporter.print_footer(result.nfev, result.chisqr, result.redchi)
    except KeyboardInterrupt:
        print_calculation_stopped_error()
    except ValueError:
        print_value_error()

    return deepcopy(minimizer.result.params)


def residuals_hierarchical(params: Parameters, param_tree: ParamTree) -> ArrayFloat:
    # print(param_tree.ids_to_fit)
    if not param_tree.branches:
        return param_tree.experiments.residuals(params)

    for param_id in param_tree.ids_to_fit:
        params[param_id].vary = False

    residuals_list: list[ArrayFloat] = []

    for branch in param_tree.branches:
        branch_results = lmfit.minimize(
            residuals_hierarchical,
            params,
            args=(branch,),
            method="leastsq",
        )
        residuals_list.append(branch_results.residual)

    residuals = np.concatenate(residuals_list, axis=0)

    for param_id in param_tree.ids_to_fit:
        params[param_id].vary = True

    return residuals


def minimize_hierarchical(experiments: Experiments, params: Parameters, fitmethod: str):
    kws = {
        "brute": {"keep": "all"},
        "basinhopping": {"niter_success": 10},
        "differential_evolution": {"polish": True, "disp": True, "init": "sobol"},
        "ampgo": {"disp": True},
    }

    reporter = Reporter()

    param_tree = create_group_tree(experiments)

    new_params = params.copy()
    for param in new_params.values():
        if param.name not in param_tree.ids_to_fit:
            param.vary = False

    minimizer = Minimizer(
        residuals_hierarchical,
        params,
        fcn_args=(param_tree,),
        iter_cb=reporter.iter_cb,
    )

    reporter.print_header()

    try:
        result = minimizer.minimize(method=fitmethod, **(kws.get(fitmethod, {})))
        reporter.print_footer(result.nfev, result.chisqr, result.redchi)
    except KeyboardInterrupt:
        print_calculation_stopped_error()
    except ValueError:
        print_value_error()

    return deepcopy(minimizer.result.params)



"""The fitting module contains the code for fitting the experimental data."""
from __future__ import annotations

from typing import TYPE_CHECKING

from rich.progress import track

from chemex.containers.experiments import Experiments, generate_exp_for_statistics
from chemex.messages import (
    print_calculation_stopped_error,
    print_fitmethod,
    print_grid_statistic_warning,
    print_group_name,
    print_minimizing,
    print_no_data,
    print_running_statistics,
    print_step_name,
    print_value_error,
)
from chemex.optimize.gridding import run_grid
from chemex.optimize.grouping import create_groups
from chemex.optimize.helper import (
    calculate_statistics,
    execute_post_fit,
    execute_post_fit_groups,
    print_header,
    print_values_stat,
)
from chemex.optimize.minimizer import (
    minimize,
    minimize_with_report,
)
from chemex.parameters import database

if TYPE_CHECKING:
    from pathlib import Path

    from chemex.configuration.methods import Methods, Statistics


def _run_statistics(
    experiments: Experiments,
    path: Path,
    fitmethod: str | None = None,
    statistics: Statistics | None = None,
) -> None:
    if statistics is None:
        return

    methods = {
        "mc": {"message": "Monte Carlo", "filename": "monte_carlo.out"},
        "bs": {"message": "bootstrap", "filename": "bootstrap.out"},
        "bsn": {"message": "nucleus-based bootstrap", "filename": "bootstrap_ns.out"},
    }

    params_lf = database.build_lmfit_params(experiments.param_ids)
    ids_vary = [param.name for param in params_lf.values() if param.vary]

    for statistic_name, iter_nb in statistics.model_dump().items():
        if iter_nb is None:
            continue

        method = methods[statistic_name]

        print_running_statistics(method["message"])

        with (path / method["filename"]).open(mode="w") as fileout:
            fileout.write(print_header(ids_vary))

            try:
                for _ in track(range(iter_nb), total=iter_nb, description="   "):
                    exp_stat = generate_exp_for_statistics(experiments, statistic_name)
                    params_lf = database.build_lmfit_params(exp_stat.param_ids)
                    params_fit = minimize(exp_stat, params_lf, fitmethod)
                    stats = calculate_statistics(exp_stat, params_fit)
                    chisqr = stats.get("chisqr", 1e32)
                    fileout.write(print_values_stat(params_fit, ids_vary, chisqr))
            except KeyboardInterrupt:
                print_calculation_stopped_error()
            except ValueError:
                print_value_error()
            finally:
                fileout.flush()


def _fit_groups(
    experiments: Experiments,
    path: Path,
    plot: str,
    fitmethod: str,
    statistics: Statistics | None,
) -> None:
    groups = create_groups(experiments)

    plot_flg = (plot == "normal" and len(groups) == 1) or plot == "all"

    print_minimizing()

    for group in groups:
        group_lmfit_params = database.build_lmfit_params(group.experiments.param_ids)
        group_path = path / group.path

        if message := group.message:
            print_group_name(message)

        best_lmfit_params = minimize_with_report(
            group.experiments,
            group_lmfit_params,
            fitmethod,
        )
        # best_lmfit_params = minimize_hierarchical(
        #     group.experiments, group_lmfit_params, fitmethod
        # )

        database.update_from_parameters(best_lmfit_params)
        execute_post_fit(group.experiments, group_path, plot_flg)

        # Run Monte Carlo and/or bootstrap analysis
        _run_statistics(
            group.experiments,
            group_path,
            fitmethod,
            statistics,
        )

    if len(groups) > 1:
        execute_post_fit_groups(experiments, path, plot)


def run_methods(
    experiments: Experiments,
    methods: Methods,
    path: Path,
    plot_level: str,
) -> None:
    for index, (section, method) in enumerate(methods.items(), start=1):
        if section:
            print_step_name(section, index, len(methods))

        # Select a subset of profiles based on "INCLUDE" and "EXCLUDE"
        experiments.select(method.selection)

        if not experiments:
            print_no_data()
            continue

        print_fitmethod(method.fitmethod)

        # Update the parameter "vary" and "expr" status
        database.set_parameter_status(method)

        if method.grid and method.statistics:
            print_grid_statistic_warning()
            method.statistics = None

        path_sect = path / section if len(methods) > 1 else path

        if method.grid:
            try:
                run_grid(
                    experiments,
                    method.grid,
                    path_sect,
                    plot_level,
                    method.fitmethod,
                )
            except KeyboardInterrupt:
                print_calculation_stopped_error()
        else:
            _fit_groups(
                experiments,
                path_sect,
                plot_level,
                method.fitmethod,
                method.statistics,
            )



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING

import numpy as np
from scipy.optimize import root

from chemex.models.constraints import pop_3st
from chemex.models.factory import model_factory
from chemex.parameters.setting import NameSetting, ParamLocalSetting
from chemex.parameters.userfunctions import user_function_registry

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions
    from chemex.typing import ArrayFloat

NAME = "3st_induced_fit"

TPL = ("temperature", "p_total", "l_total")


def calculate_residuals(
    populations: ArrayFloat,
    p_total: float,
    l_total: float,
    kd_ab: float,
    kbc: float,
    kcb: float,
) -> ArrayFloat:
    pfree, pl1, pl2 = populations
    lfree = l_total - p_total + pfree
    return np.array(
        [
            pfree + pl1 + pl2 - p_total,
            lfree * pfree - kd_ab * pl1,
            kbc * pl1 - kcb * pl2,
        ],
    )


@lru_cache(maxsize=100)
def calculate_concentrations(
    p_total: float,
    l_total: float,
    kd_ab: float,
    kbc: float,
    kcb: float,
) -> dict[str, float]:
    concentrations_start = (p_total, 0.0, 0.0)
    results = root(
        calculate_residuals,
        concentrations_start,
        args=(p_total, l_total, kd_ab, kbc, kcb),
    )
    return dict(zip(("pfree", "pl1", "pl2"), results["x"], strict=True))


def make_settings_3st_induced_fit(
    conditions: Conditions,
) -> dict[str, ParamLocalSetting]:
    p_total = conditions.p_total
    l_total = conditions.l_total
    if p_total is None:
        msg = f"'p_total' must be specified to use the '{NAME}' model"
        raise ValueError(msg)
    if l_total is None:
        msg = f"'l_total' must be specified to use the '{NAME}' model"
        raise ValueError(msg)
    return {
        "koff_ab": ParamLocalSetting(
            name_setting=NameSetting("koff_ab", "", ("temperature",)),
            value=100.0,
            min=0.0,
            vary=True,
        ),
        "kd_ab": ParamLocalSetting(
            name_setting=NameSetting("kd_ab", "", ("temperature",)),
            value=1e-3,
            min=0.0,
            vary=True,
        ),
        "kbc": ParamLocalSetting(
            name_setting=NameSetting("kbc", "", ("temperature",)),
            value=100.0,
            min=0.0,
            vary=True,
        ),
        "kcb": ParamLocalSetting(
            name_setting=NameSetting("kcb", "", ("temperature",)),
            value=100.0,
            min=0.0,
            vary=True,
        ),
        "kon": ParamLocalSetting(
            name_setting=NameSetting("kon", "", ("temperature",)),
            expr="{koff_ab} / max({kd_ab}, 1e-100)",
        ),
        "pfree": ParamLocalSetting(
            name_setting=NameSetting("pfree", "", TPL),
            expr=(f"calc_conc({p_total},{l_total},{{kd_ab}},{{kbc}},{{kcb}})['pfree']"),
        ),
        "l_free": ParamLocalSetting(
            name_setting=NameSetting("l_free", "", TPL),
            expr=f"{l_total} - {p_total} + {{pfree}}",
        ),
        "kab": ParamLocalSetting(
            name_setting=NameSetting("kab", "", TPL),
            expr="{kon} * {l_free}",
        ),
        "kba": ParamLocalSetting(
            name_setting=NameSetting("kba", "", TPL),
            expr="{koff_ab}",
        ),
        "pa": ParamLocalSetting(
            name_setting=NameSetting("pa", "", TPL),
            expr="pop_3st({kab}, {kba}, 0.0, 0.0, {kbc}, {kcb})['pa']",
        ),
        "pb": ParamLocalSetting(
            name_setting=NameSetting("pb", "", TPL),
            expr="pop_3st({kab}, {kba}, 0.0, 0.0, {kbc}, {kcb})['pb']",
        ),
        "pc": ParamLocalSetting(
            name_setting=NameSetting("pc", "", TPL),
            expr="pop_3st({kab}, {kba}, 0.0, 0.0, {kbc}, {kcb})['pc']",
        ),
    }


def register() -> None:
    model_factory.register(name=NAME, setting_maker=make_settings_3st_induced_fit)
    user_functions = {
        "calc_conc": calculate_concentrations,
        "pop_3st": pop_3st,
    }
    user_function_registry.register(name=NAME, user_functions=user_functions)



from __future__ import annotations

from contextlib import ExitStack
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Generic, Protocol, TypeVar

import numpy as np
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.ticker import MaxNLocator

from chemex.containers.data import Data
from chemex.messages import print_plot_filename
from chemex.plotters.plot import get_grid, plot_profile
from chemex.printers.plot import PlotPrinter, data_plot_printers

if TYPE_CHECKING:
    from pathlib import Path

    from matplotlib.axes import Axes

    from chemex.containers.profile import Profile
    from chemex.nmr.spectrometer import Spectrometer
    from chemex.typing import ArrayFloat

_GREY400 = "#BDBDBD"
_LSTYLES = ("-", "--", "-.", ":")


class CestExperimentSettings(Protocol):
    sw: float


class CestExperimentConfig(Protocol):
    experiment: CestExperimentSettings


T = TypeVar("T", bound=CestExperimentConfig)


def estimate_sigma(values: ArrayFloat) -> float:
    """Estimates standard deviation using median to exclude outliers.

    Up to 50% can be bad.

    Reference:
    Rousseeuw, Peter & Croux, Christophe. (1993). Alternatives to Median Absolute
    Deviation. Journal of the American Statistical Association. 88. 1273 - 1283.
    10.1080/01621459.1993.10476408.

    """
    if not all(values):
        return 0.0
    _values = values.reshape(1, -1)
    return float(1.1926 * np.median(np.median(abs(_values - _values.T), axis=0)))


@dataclass
class CircularShift:
    spectrometer: Spectrometer
    sw: float
    sw_ppm: float = field(init=False)

    def __post_init__(self) -> None:
        sw_ppm, sw_ref = self.spectrometer.offsets_to_ppms(np.array([self.sw, 0.0]))
        self.sw_ppm = sw_ppm - sw_ref

    def centre(self, array: ArrayFloat, position: float) -> ArrayFloat:
        if self.sw == np.inf:
            return array
        cs_min = position - self.sw_ppm / 2.0
        return (array - cs_min) % self.sw_ppm + cs_min


def add_resonance_positions(
    ax1: Axes,
    ax2: Axes,
    cs_values: ArrayFloat,
    centre: float,
    circular_shift: CircularShift,
) -> tuple[Axes, Axes]:
    kwargs2 = {"color": _GREY400, "linewidth": 0.75, "zorder": -1}
    cs_shifted = circular_shift.centre(cs_values, centre)
    for a_cs, a_cs_shifted, lstyle in zip(
        cs_values,
        cs_shifted,
        _LSTYLES,
        strict=False,
    ):
        ax1.axvline(a_cs_shifted, linestyle=lstyle, **kwargs2)
        ax2.axvline(a_cs_shifted, linestyle=lstyle, **kwargs2)
        if a_cs != a_cs_shifted:
            x, _ = ax2.transLimits.transform((a_cs_shifted, 0))
            ax2.text(x - 0.02, 0.95, "*", transform=ax2.transAxes)
    return ax1, ax2


def plot_dcest(
    file_pdf: PdfPages,
    name: str,
    data_exp: Data,
    data_calc: Data,
    cs_values: ArrayFloat,
    circular_shift: CircularShift,
):
    residuals = data_exp.exp - data_exp.calc
    sigma = estimate_sigma(residuals)
    centre = float(np.mean(cs_values))
    data_exp.metadata = circular_shift.centre(data_exp.metadata, centre)
    data_calc.metadata = circular_shift.centre(data_calc.metadata, centre)
    data_exp.sort()
    data_calc.sort()
    fig = plot_profile(name, data_exp, data_calc)
    ax1, ax2 = fig.axes
    xmin, xmax = sorted(ax2.get_xlim())
    ax1.set_xlim(xmax, xmin)
    ax2.set_xlim(xmax, xmin)
    ax1.xaxis.set_major_locator(MaxNLocator(6))
    ax2.xaxis.set_major_locator(MaxNLocator(6))
    ax2.set_xlabel(r"$B_1$ position (ppm)")
    ax2.set_ylabel(r"$I/I_0$")
    kwargs1 = {"facecolor": (0, 0, 0, 0.1), "edgecolor": "none"}
    ax1.fill_between(ax1.get_xlim(), -1.0 * sigma, 1.0 * sigma, **kwargs1)
    ax1.fill_between(ax1.get_xlim(), -2.0 * sigma, 2.0 * sigma, **kwargs1)

    ax1, ax2 = add_resonance_positions(ax1, ax2, cs_values, centre, circular_shift)

    file_pdf.savefig(fig)


def create_plot_data_exp(profile: Profile) -> Data:
    spectrometer = profile.spectrometer
    data = profile.data
    refs = data.refs

    ppms = spectrometer.offsets_to_ppms(data.metadata[~refs])

    intensity0 = np.mean(data.exp[refs])
    intensities = data.exp[~refs] / intensity0
    intensities_calc = data.calc[~refs] / intensity0

    errors = data.err[~refs] / abs(intensity0)
    errorbars = np.array([errors, errors]).transpose()

    data_exp = Data(exp=intensities, err=errorbars, metadata=ppms)
    data_exp.calc = intensities_calc
    data_exp.mask = data.mask[~refs]
    data_exp.sort()

    return data_exp


def create_plot_data_calc(profile: Profile) -> Data:
    spectrometer = profile.spectrometer
    data = profile.data
    refs = data.refs

    offsets_plot = get_grid(data.metadata[~refs], 400, 0.02)

    filler = np.full_like(offsets_plot, 0.0)

    data_for_calculation = Data(exp=filler, err=filler, metadata=offsets_plot)
    scale = profile.data.scale / np.mean(data.exp[refs])
    data_for_calculation.calc = scale * profile.pulse_sequence.calculate(
        spectrometer,
        data_for_calculation,
    )

    ppms = spectrometer.offsets_to_ppms(offsets_plot)
    data_fit = Data(exp=filler, err=filler, metadata=ppms)
    data_fit.calc = data_for_calculation.calc

    return data_fit


def get_state_positions(spectrometer: Spectrometer) -> ArrayFloat:
    names = (f"cs_i_{state}" for state in "abcd")
    return np.array(
        [
            spectrometer.par_values[name]
            for name in names
            if name in spectrometer.par_values
        ],
    )


class CestPlotter(Generic[T]):
    def __init__(self, filename: Path, config: T, **_extra: Any) -> None:
        self.filename = filename
        self.config = config
        self.printer: PlotPrinter = data_plot_printers["cest"]

    def _plot_profile(
        self,
        pdf: PdfPages,
        profile: Profile,
        data_exp: Data,
        data_calc: Data,
    ) -> None:
        spectrometer = profile.spectrometer
        sw = self.config.experiment.sw
        dip_positions = get_state_positions(spectrometer)
        circular_shift = CircularShift(spectrometer, sw)
        plot_dcest(
            pdf,
            str(profile.spin_system),
            data_exp,
            data_calc,
            dip_positions,
            circular_shift,
        )

    def plot(self, path: Path, profiles: list[Profile]) -> None:
        basename = path / self.filename.name
        name_pdf = basename.with_suffix(".pdf")
        name_exp = basename.with_suffix(".exp")
        name_fit = basename.with_suffix(".fit")

        print_plot_filename(name_pdf)

        with ExitStack() as stack:
            file_pdf = stack.enter_context(PdfPages(str(name_pdf)))
            file_calc = stack.enter_context(name_fit.open("w"))
            file_exp = stack.enter_context(name_exp.open("w"))
            for profile in sorted(profiles):
                data_exp = create_plot_data_exp(profile)
                data_calc = create_plot_data_calc(profile)
                self._plot_profile(file_pdf, profile, data_exp, data_calc)
                file_exp.write(
                    self.printer.print_exp(str(profile.spin_system), data_exp)
                )
                file_calc.write(
                    self.printer.print_calc(str(profile.spin_system), data_calc)
                )

    def plot_simulation(self, path: Path, profiles: list[Profile]) -> None:
        basename = path / self.filename.name
        name_pdf = basename.with_suffix(".pdf")
        name_sim = basename.with_suffix(".sim")

        print_plot_filename(name_pdf, extra=False)

        with ExitStack() as stack:
            file_pdf = stack.enter_context(PdfPages(str(name_pdf)))
            file_sim = stack.enter_context(name_sim.open("w"))
            for profile in sorted(profiles):
                data_exp = create_plot_data_exp(profile)
                data_calc = create_plot_data_calc(profile)
                self._plot_profile(file_pdf, profile, data_exp, data_calc)
                file_sim.write(
                    self.printer.print_calc(str(profile.spin_system), data_calc)
                )



"""Pace where user functions used by lmfit are stored."""

from __future__ import annotations

from collections.abc import Callable
from typing import Any, ClassVar

from chemex.configuration.conditions import Conditions
from chemex.parameters.setting import ParamLocalSetting

SettingsType = dict[str, ParamLocalSetting]
SettingMakerType = Callable[[Conditions], SettingsType]


class Registry:
    """Registry for storing all the user functions for lmfit constraints."""

    user_function_registry: ClassVar[dict[str, Any]] = {}

    def register(self, name: str, user_functions: dict[str, Callable[[Any], Any]]):
        """Register a new set of user functions."""
        self.user_function_registry[name] = user_functions
        """Register a new set of user functions."""
        self.user_function_registry[name] = user_functions

    def get(self, name: str) -> dict[str, Callable[[Any], Any]]:
        if name not in self.user_function_registry:
            return {}
        return self.user_function_registry[name]


user_function_registry = Registry()



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from chemex.containers.data import Data


class PlotPrinter(Protocol):
    def print_exp(self, name: str, data: Data) -> str:
        ...

    def print_calc(self, name: str, data: Data) -> str:
        ...


@dataclass
class DataPlotPrinter:
    first_column_name: str = ""
    first_column_fmt: str = "12.2f"

    def print_exp(self, name: str, data: Data) -> str:
        output = [
            f"[{name}]",
            (
                f"# {self.first_column_name:>12s} {'INTENSITY (EXP)':>17s}"
                f" {'ERROR (EXP)':>17s}"
            ),
        ]

        for metadata, exp, err, mask in zip(
            data.metadata,
            data.exp,
            data.err,
            data.mask,
            strict=True,
        ):
            start = " " if mask else "#"
            end = "" if mask else " # NOT USED IN THE FIT"
            output.append(
                f"{start} {metadata: {self.first_column_fmt}} {exp: 17.8e}"
                f" {err[0]: 17.8e}{end}",
            )
        return "\n".join(output) + "\n\n"

    def print_calc(self, name: str, data: Data) -> str:
        output = [
            f"[{name}]",
            f"# {self.first_column_name:>12s} {'INTENSITY (CALC)':>17s}",
        ]

        for metadata, calc in zip(data.metadata, data.calc, strict=True):
            output.append(f"  {metadata: {self.first_column_fmt}} {calc: 17.8e}")

        return "\n".join(output) + "\n\n"


@dataclass
class CpmgDataPlotPrinter(DataPlotPrinter):
    first_column_name: str = "NU_CPMG (Hz)"
    first_column_fmt: str = "12.2f"

    def print_exp(self, name: str, data: Data) -> str:
        output = [
            f"[{name}]",
            (
                f"# {self.first_column_name:>12s} {'R2 (EXP)':>17s}"
                f" {'ERROR DOWN (EXP)':>17s} {'ERROR UP (EXP)':>17s}"
            ),
        ]

        for metadata, exp, err, mask in zip(
            data.metadata,
            data.exp,
            data.err,
            data.mask,
            strict=True,
        ):
            start = " " if mask else "#"
            end = "" if mask else " # NOT USED IN THE FIT"
            output.append(
                f"{start} {metadata: {self.first_column_fmt}} {exp: 17.8e}"
                f" {err[0]: 17.8e} {err[1]: 17.8e}{end}",
            )
        return "\n".join(output) + "\n\n"

    def print_calc(self, name: str, data: Data) -> str:
        output = [f"[{name}]", f"# {self.first_column_name:>12s} {'R2 (CALC)':>17s}"]

        for metadata, calc in zip(data.metadata, data.calc, strict=True):
            output.append(f"  {metadata: {self.first_column_fmt}} {calc: 17.8e}")

        return "\n".join(output) + "\n\n"


data_plot_printers: dict[str, PlotPrinter] = {
    "cest": DataPlotPrinter("OFFSET (PPM)", "12.2f"),
    "cpmg": CpmgDataPlotPrinter(),
    "relaxation": DataPlotPrinter("TIME (S)", "12.6f"),
}



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from scipy import stats

from chemex.messages import (
    print_chi2,
    print_group_name,
    print_making_plots,
    print_plotting_canceled,
    print_writing_results,
)
from chemex.parameters import database
from chemex.printers.parameters import write_parameters

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

    from lmfit import Parameters as ParametersLF

    from chemex.containers.experiments import Experiments


def calculate_statistics(
    experiments: Experiments,
    params_lf: ParametersLF,
) -> dict[str, int | float]:
    residuals = experiments.residuals(params_lf)
    ndata = len(residuals)
    nvarys = len(
        [param for param in params_lf.values() if param.vary and not param.expr],
    )
    chisqr = sum(residuals**2)
    redchi = chisqr / max(1, ndata - nvarys)
    _neg2_log_likel = ndata * np.log(chisqr / ndata)
    aic = _neg2_log_likel + 2 * nvarys
    bic = _neg2_log_likel + np.log(ndata) * nvarys
    _, ks_p_value = stats.kstest(residuals, "norm")
    pvalue: float = 1.0 - stats.chi2.cdf(chisqr, ndata - nvarys)
    return {
        "ndata": ndata,
        "nvarys": nvarys,
        "chisqr": chisqr,
        "redchi": redchi,
        "pvalue": pvalue,
        "ks_pvalue": ks_p_value,
        "aic": aic,
        "bic": bic,
    }


def _write_statistics(experiments: Experiments, path: Path):
    """Write fitting statistics to a file."""
    params_lf = database.build_lmfit_params(experiments.param_ids)
    stats = calculate_statistics(experiments, params_lf)
    filename = path / "statistics.toml"
    with filename.open(mode="w") as f:
        f.write(f"\"number of data points\"                = {stats['ndata']}\n")
        f.write(f"\"number of variables\"                  = {stats['nvarys']}\n")
        f.write(f"\"chi-square\"                           = {stats['chisqr']: .5e}\n")
        f.write(f"\"reduced-chi-square\"                   = {stats['redchi']: .5e}\n")
        f.write(f"\"chi-squared test\"                     = {stats['pvalue']: .5e}\n")
        f.write(
            f"\"Kolmogorov-Smirnov test\"              = {stats['ks_pvalue']: .5e}\n",
        )
        f.write(f"\"Akaike Information Criterion (AIC)\"   = {stats['aic']: .5e}\n")
        f.write(f"\"Bayesian Information Criterion (BIC)\" = {stats['bic']: .5e}\n")


def _write_files(experiments: Experiments, path: Path):
    """Write the results of the fit to output files."""
    print_writing_results(path)
    path.mkdir(parents=True, exist_ok=True)
    write_parameters(experiments, path)
    experiments.write(path)
    _write_statistics(experiments, path=path)


def _write_simulation_files(experiments: Experiments, path: Path):
    """Write the results of the simulation to output files."""
    print_writing_results(path)
    path.mkdir(parents=True, exist_ok=True)
    write_parameters(experiments, path)
    experiments.write(path)


def _write_plots(experiments: Experiments, path: Path):
    """Plot the experimental and fitted data."""
    print_making_plots()

    path_ = path / "Plots"
    path_.mkdir(parents=True, exist_ok=True)
    try:
        experiments.plot(path=path_)
    except KeyboardInterrupt:
        print_plotting_canceled()


def _write_simulation_plots(experiments: Experiments, path: Path):
    """Plot the experimental and fitted data."""
    print_making_plots()

    path_ = path / "Plots"
    path_.mkdir(parents=True, exist_ok=True)
    try:
        experiments.plot_simulation(path=path_)
    except KeyboardInterrupt:
        print_plotting_canceled()


def execute_post_fit(
    experiments: Experiments,
    path: Path,
    plot: bool = False,
):
    _write_files(experiments, path)
    if plot:
        _write_plots(experiments, path)


def execute_simulation(
    experiments: Experiments,
    path: Path,
    plot: bool = False,
):
    experiments.prepare_for_simulation()
    _write_simulation_files(experiments, path)
    if plot:
        _write_simulation_plots(experiments, path)


def execute_post_fit_groups(experiments: Experiments, path: Path, plot: str) -> None:
    print_group_name("All groups")
    params_lf = database.build_lmfit_params(experiments.param_ids)
    statistics = calculate_statistics(experiments, params_lf)
    print_chi2(statistics["chisqr"], statistics["redchi"])
    execute_post_fit(experiments, path / "All", plot != "nothing")


def print_header(grid: Iterable[str]) -> str:
    parameters = database.get_parameters(grid)
    header_pnames = " ".join(f"{parameters[param_id].param_name}" for param_id in grid)
    return f"# {header_pnames} [χ²]\n"


def print_values(values: Iterable[float], chisqr: float) -> str:
    body_values = " ".join(f"{value:.5e}" for value in values)
    return f"  {body_values} {chisqr:.5e}\n"


def print_values_stat(
    params_lf: ParametersLF,
    fnames: Iterable[str],
    chisqr: float,
) -> str:
    body_values_list = []
    for fname in fnames:
        if fname in params_lf:
            body_values_list.append(f"{params_lf[fname].value:12.5e}")
        else:
            body_values_list.append(f"{'--':^12s}")
    body_values = " ".join(body_values_list)
    return f"  {body_values} {chisqr:.5e}\n"



from __future__ import annotations

from functools import reduce
from itertools import product
from typing import TYPE_CHECKING

import numpy as np
from cachetools import cached
from cachetools.keys import hashkey
from scipy.linalg import expm

if TYPE_CHECKING:
    # Imports related to type checking and annotations
    from collections.abc import Iterable, Sequence

    from chemex.nmr.constants import Distribution
    from chemex.nmr.liouvillian import LiouvillianIS
    from chemex.typing import ArrayFloat

    DictArrayFloat = dict[str, ArrayFloat]

# A small value used for numerical stability
SMALL_VALUE = 1e-6


def calculate_propagators(
    liouv: ArrayFloat,
    delays: float | Iterable[float],
    dephasing: bool = False,
) -> ArrayFloat:
    """Calculate the propagators for the given delays.

    Parameters
    ----------
    liouv : ArrayFloat
        The Liouvillians of the system
    delays : float | Iterable[float]
        The delays to calculate the propagators for
    dephasing : bool, optional
        Whether to apply dephasing, by default False

    Returns:
    -------
    ArrayFloat
        The propagators
    """
    # Ensure delays is a 1D NumPy array
    delays = np.asarray(delays).flatten()

    if delays.size == 1 and not dephasing:
        # If there's only one delay and no dephasing, calculate the propagator directly
        # using scipy expm function
        return expm(liouv * delays[0])

    # Calculate eigenvalues and eigenvectors of the Liouvillian matrices
    eigenvalues, eigenvectors = np.linalg.eig(liouv)

    if dephasing:
        # If dephasing is requested, adjust eigenvalues
        eigenvalues = np.where(
            abs(eigenvalues.imag) < SMALL_VALUE,
            eigenvalues,
            eigenvalues * 1e9,
        )

    # Calculate the exponentiated eigenvalues for each delay
    evt = np.exp(np.multiply.outer(delays, eigenvalues))

    # Create an empty array for diagonal matrices
    diag = np.empty((*delays.shape, *liouv.shape), dtype=np.complex128)

    # Calculate diagonal matrices for each delay
    for index in product(*(range(x) for x in evt.shape[:-1])):
        diag[index] = np.diag(evt[index])

    # Calculate the propagators using matrix multiplication
    propagators = eigenvectors @ diag @ np.linalg.inv(eigenvectors)

    if propagators.shape[0] == 1:
        # If there's only one propagator, return it with no extra time dimension
        propagators = propagators[0]

    # Return the real part of the propagators
    return propagators.real


def _get_key(liouvillian: LiouvillianIS, *args, **kwargs):
    return hashkey(liouvillian.basis, *args, **kwargs)


@cached(cache={}, key=_get_key)
def _make_perfect180(liouvillian: LiouvillianIS, spin: str) -> ArrayFloat:
    size = liouvillian.size
    identity = np.eye(size).reshape((1, 1, size, size))
    compx, compy, compz = (f"{spin}{axis}" for axis in "xyz")
    perfect180 = {comp: identity.copy() for comp in (compx, compy)}
    for comp in liouvillian.basis.components:
        vect = liouvillian.vectors[comp].ravel()
        if compx in comp or compz in comp:
            perfect180[compy] -= 2 * np.diag(vect)
        if compy in comp or compz in comp:
            perfect180[compx] -= 2 * np.diag(vect)
    p180 = [perfect180[comp] for comp in (compx, compy, compx, compy)]
    return np.array(p180)


@cached(cache={}, key=_get_key)
def _make_perfect90(liouvillian: LiouvillianIS, spin: str) -> ArrayFloat:
    size = liouvillian.size
    zeros = np.zeros((size, size))
    rot = liouvillian.matrices.get(f"b1x_{spin}", zeros)
    return np.asarray(expm(0.25 * rot)).reshape((1, 1, size, size))


@cached(cache={}, key=_get_key)
def _get_phases(liouvillian: LiouvillianIS) -> DictArrayFloat:
    phases = {}
    size = liouvillian.size
    zeros = np.zeros((size, size))
    for spin in "is":
        l_rotz = liouvillian.matrices.get(f"rotz_{spin}", zeros)
        phases[spin] = np.array([expm(n * 0.5 * np.pi * l_rotz) for n in range(4)])
    return phases


class Spectrometer:
    def _add_phases(self, propagator: ArrayFloat, spin: str = "i") -> ArrayFloat:
        phases = self._phases[spin]
        return np.array([phases[i] @ propagator @ phases[-i] for i in range(4)])

    def __init__(self, liouvillian: LiouvillianIS) -> None:
        self.liouvillian = liouvillian
        size = self.liouvillian.size
        self.identity = np.eye(self.liouvillian.size).reshape((1, 1, size, size))
        self._phases = _get_phases(liouvillian)
        self.perfect90_i = self._add_phases(_make_perfect90(liouvillian, "i"))
        self.perfect180_i = _make_perfect180(liouvillian, "i")
        self.perfect180_s = _make_perfect180(liouvillian, "s")
        self.zfilter = self.keep(
            self.identity,
            components={"ie", "se", "iz", "sz", "2izsz"}
            & set(liouvillian.basis.components),
        )
        self.calculate_i_flag = True
        self.calculate_s_flag = True
        self._pw90_i = 0.0
        self._pw90_s = 0.0
        self._p90_i = np.array(0.0)
        self._p180_i = np.array(0.0)
        self._p240_i = np.array(0.0)
        self._p180_s = np.array(0.0)

    def keep(self, magnetization: ArrayFloat, components: Iterable[str]) -> ArrayFloat:
        return self.liouvillian.keep(magnetization, components)

    def update(self, par_values: dict[str, float]) -> None:
        self.liouvillian.update(par_values)
        self.calculate_i_flag = True
        self.calculate_s_flag = True

    @property
    def par_values(self) -> dict[str, float]:
        return self.liouvillian.par_values

    @property
    def carrier_i(self) -> float:
        return self.liouvillian.carrier_i

    @carrier_i.setter
    def carrier_i(self, value: float):
        self.liouvillian.carrier_i = value
        self.calculate_i_flag = True
        self.calculate_s_flag = True

    @property
    def carrier_s(self) -> float:
        return self.liouvillian.carrier_s

    @carrier_s.setter
    def carrier_s(self, value: float):
        self.liouvillian.carrier_s = value
        self.calculate_i_flag = True
        self.calculate_s_flag = True

    @property
    def offset_i(self) -> float:
        return self.liouvillian.offset_i

    @offset_i.setter
    def offset_i(self, value: float):
        self.liouvillian.offset_i = value
        self.calculate_i_flag = True
        self.calculate_s_flag = True

    @property
    def offset_s(self) -> float:
        return self.liouvillian.offset_s

    @offset_s.setter
    def offset_s(self, value: float):
        self.liouvillian.offset_s = value
        self.calculate_i_flag = True
        self.calculate_s_flag = True

    @property
    def b1_i(self) -> float:
        return self.liouvillian.b1_i

    @b1_i.setter
    def b1_i(self, value: float):
        self._pw90_i = 1.0 / (4.0 * value) if value else 0.0
        self.liouvillian.b1_i = value
        self.calculate_i_flag = True
        self.calculate_s_flag = True

    @property
    def b1_s(self) -> float:
        return self.liouvillian.b1_s

    @b1_s.setter
    def b1_s(self, value: float):
        self._pw90_s = 1.0 / (4.0 * value) if value else 0.0
        self.liouvillian.b1_s = value
        self.calculate_i_flag = True
        self.calculate_s_flag = True

    @property
    def b1_i_inh_scale(self) -> float:
        return self.liouvillian.b1_i_inh_scale

    @b1_i_inh_scale.setter
    def b1_i_inh_scale(self, value: float):
        self.liouvillian.b1_i_inh_scale = value
        self.calculate_i_flag = True
        self.calculate_s_flag = True

    @property
    def b1_i_inh_res(self) -> int:
        return self.liouvillian.b1_i_inh_res

    @b1_i_inh_res.setter
    def b1_i_inh_res(self, value: int):
        self.liouvillian.b1_i_inh_res = value
        self.calculate_i_flag = True
        self.calculate_s_flag = True

    @property
    def jeff_i(self) -> Distribution:
        return self.liouvillian.jeff_i

    @jeff_i.setter
    def jeff_i(self, value: Distribution):
        self.liouvillian.jeff_i = value
        self.calculate_i_flag = True
        self.calculate_s_flag = True

    def get_equilibrium(self) -> ArrayFloat:
        return self.liouvillian.get_equilibrium()

    def get_start_magnetization(
        self,
        terms: Iterable[str],
        atom: str = "h",
    ) -> ArrayFloat:
        return self.liouvillian.get_start_magnetization(terms=terms, atom=atom)

    @property
    def detection(self) -> str:
        return self.liouvillian.detection

    @detection.setter
    def detection(self, value: str):
        self.liouvillian.detection = value

    def detect(self, magnetization: ArrayFloat) -> float:
        return self.liouvillian.detect(magnetization)

    @property
    def gradient_dephasing(self) -> float:
        return self.liouvillian.gradient_dephasing

    @gradient_dephasing.setter
    def gradient_dephasing(self, value: float):
        self.liouvillian.gradient_dephasing = value
        self.calculate_i_flag = True
        self.calculate_s_flag = True

    def delays(self, times: float | Iterable[float]) -> ArrayFloat:
        return calculate_propagators(self.liouvillian.l_free, times)

    def pulse_i(
        self,
        times: float | Iterable[float],
        phase: float,
        scale: float = 1.0,
    ) -> ArrayFloat:
        dephased = self.b1_i_inh_scale == np.inf
        rad = phase * np.pi * 0.5
        liouv = (
            self.liouvillian.l_free
            + scale * np.cos(rad) * self.liouvillian.l_b1x_i
            + scale * np.sin(rad) * self.liouvillian.l_b1y_i
        )
        return calculate_propagators(liouv, times, dephased)

    def pulse_s(
        self,
        times: float | Iterable[float],
        phase: float,
        scale: float = 1.0,
    ) -> ArrayFloat:
        rad = phase * np.pi * 0.5
        liouv = (
            self.liouvillian.l_free
            + scale * np.cos(rad) * self.liouvillian.l_b1x_s
            + scale * np.sin(rad) * self.liouvillian.l_b1y_s
        )
        return calculate_propagators(liouv, times)

    def pulse_is(
        self,
        times: float | Iterable[float],
        phase_i: float,
        phase_s: float,
    ) -> ArrayFloat:
        dephased = self.b1_i_inh_scale == np.inf
        liouv = (
            self.liouvillian.l_free
            + np.cos(phase_i * np.pi * 0.5) * self.liouvillian.l_b1x_i
            + np.sin(phase_i * np.pi * 0.5) * self.liouvillian.l_b1y_i
            + np.cos(phase_s * np.pi * 0.5) * self.liouvillian.l_b1x_s
            + np.sin(phase_s * np.pi * 0.5) * self.liouvillian.l_b1y_s
        )
        return calculate_propagators(liouv, times, dephased)

    def shaped_pulse_i(
        self,
        pw: float,
        amplitudes: Sequence[float],
        phases: Iterable[float],
    ) -> ArrayFloat:
        time = pw / len(amplitudes)
        pairs = list(zip(amplitudes, phases, strict=True))
        pulses = {
            (amp, ph): self.pulse_i(time, ph, scale=amp) for amp, ph in set(pairs)
        }
        base = reduce(np.matmul, (pulses[pair] for pair in reversed(pairs)))
        return self._add_phases(base, "i")

    def _calculate_base_pulses_i(self) -> None:
        if self.calculate_i_flag:
            pws = np.array([1.0, 2.0, 8.0 / 3.0]) * self._pw90_i
            base = self.pulse_i(pws, 0.0)
            pulses = self._add_phases(base, "i")
            self._p90_i, self._p180_i, self._p240_i = pulses.swapaxes(0, 1)

    @property
    def p90_i(self) -> ArrayFloat:
        self._calculate_base_pulses_i()
        return self._p90_i

    @property
    def p180_i(self) -> ArrayFloat:
        self._calculate_base_pulses_i()
        return self._p180_i

    @property
    def p240_i(self) -> ArrayFloat:
        self._calculate_base_pulses_i()
        return self._p240_i

    @property
    def p9018090_i_1(self) -> ArrayFloat:
        return self.p90_i[[3, 0, 1, 2]] @ self.p180_i @ self.p90_i[[3, 0, 1, 2]]

    @property
    def p9018090_i_2(self) -> ArrayFloat:
        return self.p90_i[[1, 2, 3, 0]] @ self.p180_i @ self.p90_i[[1, 2, 3, 0]]

    @property
    def p9024090_i_1(self) -> ArrayFloat:
        return self.p90_i[[3, 0, 1, 2]] @ self.p240_i @ self.p90_i[[3, 0, 1, 2]]

    @property
    def p9024090_i_2(self) -> ArrayFloat:
        return self.p90_i[[1, 2, 3, 0]] @ self.p240_i @ self.p90_i[[1, 2, 3, 0]]

    def _calculate_base_pulses_s(self) -> None:
        if self.calculate_s_flag:
            pws = np.array([2.0]) * self._pw90_s
            base = self.pulse_s(pws, 0.0)
            pulses = self._add_phases(base, "s")
            self._p180_s = pulses.swapaxes(0, 1)

    @property
    def p180_s(self) -> ArrayFloat:
        self._calculate_base_pulses_s()
        return self._p180_s

    def p9024090_nh(self, reverse: bool = False) -> ArrayFloat:
        ph_n = 1 if reverse else 3
        ph_h = 3 if reverse else 1
        pw240i, pw9024090i = np.array([8.0, 14.0]) * self._pw90_i / 3.0
        pw240s, pw9024090s = np.array([8.0, 14.0]) * self._pw90_s / 3.0
        t0, t1, t2, t3 = 0.5 * np.diff(
            np.sort([pw240i, pw240s, pw9024090i, pw9024090s]),
            prepend=0.0,
        )
        p0 = self.pulse_is(2.0 * t0, 0, 0)
        if pw9024090i <= pw9024090s:
            p1 = self.pulse_is(t1, ph_n, 0)
            p2 = (
                self.pulse_is(t2, ph_n, ph_h)
                if pw9024090i > pw240s
                else self.pulse_s(t2, 0)
            )
            p3 = self.pulse_s(t3, ph_h)
        else:
            p1 = self.pulse_is(t1, 0, ph_h)
            p2 = (
                self.pulse_is(t2, ph_n, ph_h)
                if pw9024090s > pw240i
                else self.pulse_i(t2, 0)
            )
            p3 = self.pulse_i(t3, ph_n)
        pw9024090is_xx = p3 @ p2 @ p1 @ p0 @ p1 @ p2 @ p3
        return self._add_phases(self._add_phases(pw9024090is_xx, "s"), "i")

    @property
    def p9024090_nh_1(self) -> ArrayFloat:
        return self.p9024090_nh()

    @property
    def p9024090_nh_2(self) -> ArrayFloat:
        return self.p9024090_nh(reverse=True)

    def calculate_shifts(self) -> ArrayFloat:
        liouv = self.liouvillian.l_free.reshape(
            (self.liouvillian.size, self.liouvillian.size),
        )
        return np.linalg.eigvals(liouv).imag

    def offsets_to_ppms(self, offsets: ArrayFloat) -> ArrayFloat:
        return self.liouvillian.offsets_to_ppms(offsets)

    def ppms_to_offsets(self, ppms: ArrayFloat | float) -> ArrayFloat | float:
        return self.liouvillian.ppms_to_offsets(ppms)



from __future__ import annotations

from functools import lru_cache

import numpy as np
from scipy import linalg


@lru_cache(maxsize=100)
def pop_2st(kab: float = 0.0, kba: float = 0.0) -> dict[str, float]:
    if np.isclose(kab, 0.0):
        return {"pa": 1.0, "pb": 0.0}

    if np.isclose(kba, 0.0):
        return {"pa": 0.0, "pb": 1.0}

    kex = kab + kba

    return {"pa": kba / kex, "pb": kab / kex}


@lru_cache(maxsize=100)
def pop_3st(
    kab: float = 0.0,
    kba: float = 0.0,
    kac: float = 0.0,
    kca: float = 0.0,
    kbc: float = 0.0,
    kcb: float = 0.0,
) -> dict[str, float]:
    if np.isclose(kab + kba + kac + kca, 0.0):
        pb, pc = pop_2st(kbc, kcb).values()
        return {"pa": 0.0, "pb": pb, "pc": pc}

    if np.isclose(kab + kba + kbc + kcb, 0.0):
        pa, pc = pop_2st(kac, kca).values()
        return {"pa": pa, "pb": 0.0, "pc": pc}

    if np.isclose(kac + kca + kbc + kcb, 0.0):
        pa, pb = pop_2st(kab, kba).values()
        return {"pa": pa, "pb": pb, "pc": 0.0}

    mat = np.array(
        [
            [-kab - kac, kba, kca],
            [kab, -kba - kbc, kcb],
            [1.0, 1.0, 1.0],
        ],
    )

    if np.isclose(np.linalg.det(mat), 0.0):
        return {"pa": 1.0, "pb": 0.0, "pc": 0.0}

    vec = np.array([0.0, 0.0, 1.0])
    pa, pb, pc = linalg.solve(mat, vec)

    return {"pa": pa, "pb": pb, "pc": pc}


@lru_cache(maxsize=100)
def pop_4st(
    kab: float = 0.0,
    kba: float = 0.0,
    kac: float = 0.0,
    kca: float = 0.0,
    kad: float = 0.0,
    kda: float = 0.0,
    kbc: float = 0.0,
    kcb: float = 0.0,
    kbd: float = 0.0,
    kdb: float = 0.0,
    kcd: float = 0.0,
    kdc: float = 0.0,
) -> dict[str, float]:
    if np.isclose(kab + kba + kac + kca + kad + kda, 0.0):
        pb, pc, pd = pop_3st(kbc, kcb, kbd, kdb, kcd, kdc).values()
        return {"pa": 0.0, "pb": pb, "pc": pc, "pd": pd}

    if np.isclose(kab + kba + kbc + kcb + kbd + kdb, 0.0):
        pa, pc, pd = pop_3st(kac, kca, kad, kda, kcd, kdc).values()
        return {"pa": pa, "pb": 0.0, "pc": pc, "pd": pd}

    if np.isclose(kac + kca + kbc + kcb + kcd + kdc, 0.0):
        pa, pb, pd = pop_3st(kab, kba, kad, kda, kbd, kdb).values()
        return {"pa": pa, "pb": pb, "pc": 0.0, "pd": pd}

    if np.isclose(kad + kda + kbd + kdb + kcd + kdc, 0.0):
        pa, pb, pc = pop_3st(kab, kba, kac, kca, kbc, kcb).values()
        return {"pa": pa, "pb": pb, "pc": pc, "pd": 0.0}

    mat = np.array(
        [
            [-kab - kac - kad, kba, kca, kda],
            [kab, -kba - kbc - kbd, kcb, kdb],
            [kac, kbc, -kca - kcb - kcd, kdc],
            [1.0, 1.0, 1.0, 1.0],
        ],
    )

    if np.isclose(np.linalg.det(mat), 0.0):
        return {"pa": 1.0, "pb": 0.0, "pc": 0.0, "pd": 0.0}

    vec = np.array([0.0, 0.0, 0.0, 1.0])
    pa, pb, pc, pd = linalg.solve(mat, vec)

    return {"pa": pa, "pb": pb, "pc": pc, "pd": pd}



from __future__ import annotations

from functools import total_ordering
from re import search

from chemex.parameters.spin_system.atom import Atom
from chemex.parameters.spin_system.constants import STANDARD_ATOM_NAMES
from chemex.parameters.spin_system.group import Group


@total_ordering
class Spin:
    def __init__(self, name: str, group_for_completion: Group | None = None) -> None:
        self.group, self.atom = self.split_group_atom(name.strip().upper())
        if not self.group and group_for_completion:
            self.group = group_for_completion
        self.search_keys = self.group.search_keys | self.atom.search_keys

    @staticmethod
    def split_group_atom(name: str) -> tuple[Group, Atom]:
        if name == "?":
            return Group(""), Atom("")
        found_digit = search("[0-9]", name)
        first_digit = found_digit.start() if found_digit else 0
        found_atom = search("[HCNQM]", name[first_digit:])
        if not found_atom:
            if name in STANDARD_ATOM_NAMES:
                return Group(""), Atom(name)
            return Group(name), Atom("")
        atom_index = first_digit + found_atom.start()
        return Group(name[:atom_index]), Atom(name[atom_index:])

    @property
    def name(self) -> str:
        return f"{self.group}{self.atom}"

    def match(self, other: Spin) -> bool:
        return self.group.match(other.group) and self.atom.match(other.atom)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        self_tuple = (self.atom.nucleus.name, self.group, self.atom)
        other_tuple = (other.atom.nucleus.name, other.group, other.atom)
        return self_tuple < other_tuple

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.name == other.name

    def __hash__(self) -> int:
        return hash(self.name)

    def __bool__(self) -> bool:
        return bool(self.name)

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name



from __future__ import annotations

from itertools import product
from typing import TYPE_CHECKING, TypeVar

import numpy as np
from scipy.constants import hbar, mu_0

from chemex.models.model import model
from chemex.nmr.constants import GAMMA

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions
    from chemex.nmr.basis import Basis
    from chemex.typing import ArrayFloat

    # Type definition
    T = TypeVar("T", float, ArrayFloat)


def _calculate_jw(tauc: float, s2: float, w: T) -> T:
    tauc_ = tauc * 1e-9
    return 2.0 / 5.0 * tauc_ * s2 / (1.0 + (w * tauc_) ** 2)


class RatesIS:
    gi: float
    gs: float
    ris3: float
    rih3: float
    rsh3: float
    csa_i: ArrayFloat
    csa_s: ArrayFloat
    phi_i: ArrayFloat
    phi_s: ArrayFloat

    def __init__(self) -> None:
        self.gh = GAMMA["h"]

        # Dipolar factors
        self.dis = -mu_0 * hbar * self.gi * self.gs / (8.0 * np.pi * self.ris3)
        dih = -mu_0 * hbar * self.gi * self.gh / (8.0 * np.pi * self.rih3)
        dsh = -mu_0 * hbar * self.gs * self.gh / (8.0 * np.pi * self.rsh3)
        self.dis2 = self.dis * self.dis
        self.dih2 = dih * dih
        self.dsh2 = dsh * dsh

        # CSA factors
        self.delta_i = self.csa_i[:2] - self.csa_i[2]
        self.delta_s = self.csa_s[:2] - self.csa_s[2]

    def __call__(self, h_frq: float, tauc: float, s2: float) -> dict[str, float]:
        # B0 in Tesla
        b0 = 2.0 * np.pi * 1e6 * h_frq / self.gh

        # Larmor angular frequencies
        wi, ws, wh = -np.array([self.gi, self.gs, self.gh]) * b0

        # CSA factors
        ci_ = -wi * self.delta_i / 3.0
        cs_ = -ws * self.delta_s / 3.0
        ci2 = ci_[0] ** 2 + ci_[1] ** 2 - ci_[0] * ci_[1]
        cs2 = cs_[0] ** 2 + cs_[1] ** 2 - cs_[0] * cs_[1]
        p2i = 0.5 * (3.0 * np.cos(self.phi_i[:2]) ** 2 - 1.0)
        p2s = 0.5 * (3.0 * np.cos(self.phi_s[:2]) ** 2 - 1.0)
        ci = (ci_ * p2i).sum()
        cs = (cs_ * p2s).sum()

        j0, ji, js, jh, jmis, jpis, jmih, jpih, jmsh, jpsh = _calculate_jw(
            tauc,
            s2,
            np.array(
                [0.0, wi, ws, wh, wi - ws, wi + ws, wi - wh, wi + wh, ws - wh, ws + wh],
            ),
        )

        return {
            "r2_i": 0.5
            * (
                ci2 * (4 * j0 + 3 * ji)
                + self.dis2 * (4 * j0 + 3 * ji + jmis + 6 * js + 6 * jpis)
                + self.dih2 * (4 * j0 + 3 * ji + jmih + 6 * jh + 6 * jpih)
            ),
            "r2_s": 0.5
            * (
                cs2 * (4 * j0 + 3 * js)
                + self.dis2 * (4 * j0 + 3 * js + jmis + 6 * ji + 6 * jpis)
                + self.dsh2 * (4 * j0 + 3 * js + jmsh + 6 * jh + 6 * jpsh)
            ),
            "r1_i": (
                ci2 * 3 * ji
                + self.dis2 * (3 * ji + jmis + 6 * jpis)
                + self.dih2 * (3 * ji + jmih + 6 * jpih)
            ),
            "r1_s": (
                cs2 * 3 * js
                + self.dis2 * (3 * js + jmis + 6 * jpis)
                + self.dsh2 * (3 * js + jmsh + 6 * jpsh)
            ),
            "r2a_i": (
                0.5 * ci2 * (4 * j0 + 3 * ji)
                + cs2 * 3 * js
                + 0.5 * self.dis2 * (4 * j0 + 3 * ji + jmis + 6 * jpis)
                + 0.5 * self.dih2 * (4 * j0 + 3 * ji + jmih + 6 * jh + 6 * jpih)
                + self.dsh2 * (3 * js + jmsh + 6 * jpsh)
            ),
            "r2a_s": (
                0.5 * cs2 * (4 * j0 + 3 * js)
                + ci2 * 3 * ji
                + 0.5 * self.dis2 * (4 * j0 + 3 * js + jmis + 6 * jpis)
                + 0.5 * self.dsh2 * (4 * j0 + 3 * js + jmsh + 6 * jh + 6 * jpsh)
                + self.dih2 * (3 * ji + jmih + 6 * jpih)
            ),
            "r2mq_is": 0.5
            * (
                ci2 * (4 * j0 + 3 * ji)
                + cs2 * (4 * j0 + 3 * js)
                + self.dis2 * (3 * ji + jmis + 3 * js + 6 * jpis)
                + self.dih2 * (4 * j0 + 3 * ji + jmih + 6 * jh + 6 * jpih)
                + self.dsh2 * (4 * j0 + 3 * js + jmsh + 6 * jh + 6 * jpsh)
            ),
            "r1a_is": (
                ci2 * 3 * ji
                + cs2 * 3 * js
                + self.dis2 * (3 * ji + 3 * js)
                + self.dih2 * (3 * ji + jmih + 6 * jpih)
                + self.dsh2 * (3 * js + jmsh + 6 * jpsh)
            ),
            "etaxy_i": self.dis * ci * (4 * j0 + 3 * ji),
            "etaxy_s": self.dis * cs * (4 * j0 + 3 * js),
            "etaz_i": self.dis * ci * 6 * ji,
            "etaz_s": self.dis * cs * 6 * js,
            "sigma_is": self.dis2 * (-jmis + 6.0 * jpis),
            "mu_is": 0.5 * self.dis2 * (-jmis + 6.0 * jpis),
        }


class RateNH(RatesIS):
    gi = GAMMA["n"]
    gs = GAMMA["h"]
    ris3 = 1.04e-10**3
    rih3 = 1.79e-10**3
    rsh3 = 1.85e-10**3
    csa_i = np.array([69.0, 42.0, -111.0]) * 1e-6
    csa_s = np.array([5.7, 0.5, -6.2]) * 1e-6
    phi_i = np.deg2rad([109.6, 90.0, 19.6])
    phi_s = np.deg2rad([10.0, 80.0, 90.0])

    def __call__(
        self,
        h_frq: float,
        tauc: float,
        s2: float,
        khh: float = 0.0,
    ) -> dict[str, float]:
        rates = super().__call__(h_frq, tauc, s2)
        if khh == 0:
            return rates
        # Make a copy of rates before adding 'khh' due to lru_cache"
        # on 'super().__call__'
        rates = rates.copy()
        rates["r2_s"] += khh
        rates["r1_s"] += khh
        rates["r2a_i"] += khh
        rates["r2a_s"] += khh
        rates["r2mq_is"] += khh
        rates["r1a_is"] += khh
        return rates


class RateNH_D(RateNH):
    rih3 = 2.50e-10**3
    rsh3 = 2.48e-10**3


class RateHN(RatesIS):
    gi = GAMMA["h"]
    gs = GAMMA["n"]
    ris3 = 1.04e-10**3
    rih3 = 1.85e-10**3
    rsh3 = 1.79e-10**3
    csa_i = np.array([5.7, 0.5, -6.2]) * 1e-6
    csa_s = np.array([69.0, 42.0, -111.0]) * 1e-6
    phi_i = np.deg2rad([10.0, 80.0, 90.0])
    phi_s = np.deg2rad([109.6, 90.0, 19.6])

    def __call__(
        self,
        h_frq: float,
        tauc: float,
        s2: float,
        khh: float = 0.0,
    ) -> dict[str, float]:
        rates = super().__call__(h_frq, tauc, s2)
        if khh == 0:
            return rates
        # Make a copy of rates before adding 'khh' due to lru_cache
        # on 'super().__call__'
        rates = rates.copy()
        rates["r2_i"] += khh
        rates["r1_i"] += khh
        rates["r2a_i"] += khh
        rates["r2a_s"] += khh
        rates["r2mq_is"] += khh
        rates["r1a_is"] += khh
        return rates


class RateHN_D(RateHN):
    rih3 = 2.48e-10**3
    rsh3 = 2.50e-10**3


class RateCH(RatesIS):
    gi = GAMMA["c"]
    gs = GAMMA["h"]
    ris3 = 1.09e-10**3
    rih3 = 1.70e-10**3
    rsh3 = 1.85e-10**3
    csa_i = np.array([-5.3, -5.3, 10.7]) / 3.0 * 1e-6
    csa_s = np.array([0.0, 0.0, 0.0]) * 1e-6
    phi_i = np.deg2rad([109.6, 90.0, 19.6])
    phi_s = np.deg2rad([0.0, 0.0, 0.0])


class RateCH_D(RateCH):
    rih3 = 2.03e-10**3
    rsh3 = 2.52e-10**3


class RateHC(RatesIS):
    gi = GAMMA["h"]
    gs = GAMMA["c"]
    ris3 = 1.09e-10**3
    rih3 = 1.85e-10**3
    rsh3 = 1.70e-10**3
    csa_i = np.array([0.0, 0.0, 0.0]) * 1e-6
    csa_s = np.array([-5.3, -5.3, 10.7]) / 3.0 * 1e-6
    phi_i = np.deg2rad([0.0, 0.0, 0.0])
    phi_s = np.deg2rad([109.6, 90.0, 19.6])


class RateHC_D(RateHC):
    rih3 = 2.52e-10**3
    rsh3 = 2.03e-10**3


rate_functions: dict[str, RatesIS] = {
    "nh": RateNH(),
    "nh_d": RateNH_D(),
    "hn": RateHN(),
    "hn_d": RateHN_D(),
    "ch": RateCH(),
    "ch_d": RateCH_D(),
    "hc": RateHC(),
    "hc_d": RateHC_D(),
}

_RATE_NAMES = [
    "r2_i",
    "r2_s",
    "r1_i",
    "r1_s",
    "r2a_i",
    "r2a_s",
    "r2mq_is",
    "r1a_is",
    "etaxy_i",
    "etaxy_s",
    "etaz_i",
    "etaz_s",
    "sigma_is",
    "mu_is",
]


def get_model_free_expressions(basis: Basis, conditions: Conditions) -> dict[str, str]:
    """It takes a basis and a set of conditions,
    and returns a dictionary of rate expressions.

    Parameters
    ----------
    basis : Basis
        The basis set to use for the model-free analysis.
    conditions : Conditions
        The conditions for the fitting.

    Returns:
    -------
        A dictionary of rate names and their corresponding expressions.

    """
    deuterated_extension = "_d" if conditions.is_deuterated else ""
    rate_function_name = f"{basis.spin_system}{deuterated_extension}"

    h_frq_str = f"{conditions.h_larmor_frq}"
    has_h_exchange = basis.spin_system in {"nh", "hn"}

    model_free_expr: dict[str, str] = {}
    for state, name in product(model.states, _RATE_NAMES):
        rate_name = f"{name}_{state}"
        khh = f", {{khh_{state}}}" if has_h_exchange else ""
        arguments = f"{h_frq_str}, {{tauc_{state}}}, {{s2_{state}}}{khh}"
        expr = f"{rate_function_name}({arguments})['{name}']"
        model_free_expr[rate_name] = expr

    return model_free_expr



from __future__ import annotations

from typing import TYPE_CHECKING

from chemex.models.factory import model_factory
from chemex.parameters.setting import NameSetting, ParamLocalSetting

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions

NAME = "1st"


def make_settings_1st(_conditions: Conditions) -> dict[str, ParamLocalSetting]:
    return {"pa": ParamLocalSetting(name_setting=NameSetting("pa"), expr="1.0")}


def register() -> None:
    model_factory.register(name=NAME, setting_maker=make_settings_1st)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Generic, TypeVar

import numpy as np
from pydantic import BaseModel, ConfigDict

from chemex.configuration.base import BaseModelLowerCase
from chemex.configuration.conditions import ConditionsFromFile


@dataclass
class ToBeFitted:
    rates: list[str] = field(default_factory=list)
    model_free: list[str] = field(default_factory=list)


class ExperimentNameSettings(BaseModelLowerCase):
    model_config = ConfigDict(str_to_lower=True)
    name: str


class RelaxationSettings(BaseModelLowerCase):
    model_config = ConfigDict(str_to_lower=True)
    name: str


class CpmgSettings(BaseModelLowerCase):
    model_config = ConfigDict(str_to_lower=True)
    name: str
    even_ncycs: bool = False


class CpmgSettingsEvenNcycs(CpmgSettings):
    even_ncycs: bool = True


class CestSettings(BaseModelLowerCase):
    model_config = ConfigDict(str_to_lower=True)
    name: str
    sw: float = np.inf


class ShiftSettings(BaseModelLowerCase):
    model_config = ConfigDict(str_to_lower=True)
    name: str


class ExperimentNameConfig(BaseModel):
    experiment: ExperimentNameSettings


ExperimentSettingsType = TypeVar("ExperimentSettingsType", bound=BaseModelLowerCase)
DataSettingsType = TypeVar("DataSettingsType", bound=BaseModelLowerCase)


class ExperimentConfig(BaseModel, Generic[ExperimentSettingsType, DataSettingsType]):
    experiment: ExperimentSettingsType
    conditions: ConditionsFromFile
    data: DataSettingsType

    @property
    def to_be_fitted(self) -> ToBeFitted:
        return ToBeFitted()



from __future__ import annotations

from itertools import permutations
from typing import TYPE_CHECKING

from chemex.models.factory import model_factory
from chemex.parameters.setting import NameSetting, ParamLocalSetting

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions


TPL = ("temperature", "p_total", "l_total")


def kex_p_to_k(states: str) -> str:
    kex = f"{{kex_{min(states)}{max(states)}}}"
    p1 = f"{{p{states[0]}}}"
    p2 = f"{{p{states[1]}}}"
    return f"{kex} * {p2} / max({p1} + {p2}, 1e-100)"


def create_kij_settings(states: str = "abc") -> dict[str, ParamLocalSetting]:
    if len(states) <= 1:
        msg = "'state_nb' should be larger than 2"
        raise ValueError(msg)
    return {
        f"k{i}{j}": ParamLocalSetting(
            name_setting=NameSetting(f"k{i}{j}", "", TPL),
            expr=kex_p_to_k(f"{i}{j}"),
        )
        for i, j in permutations(states, 2)
    }


def make_settings_3st_linear(_conditions: Conditions) -> dict[str, ParamLocalSetting]:
    kij_settings = create_kij_settings("abc")
    del kij_settings["kac"]
    del kij_settings["kca"]
    return {
        "pb": ParamLocalSetting(
            name_setting=NameSetting("pb", "", TPL),
            value=0.02,
            min=0.0,
            max=1.0,
            vary=True,
        ),
        "pc": ParamLocalSetting(
            name_setting=NameSetting("pc", "", TPL),
            value=0.02,
            min=0.0,
            max=1.0,
            vary=True,
        ),
        "kex_ab": ParamLocalSetting(
            name_setting=NameSetting("kex_ab", "", TPL),
            min=0.0,
            value=200.0,
            vary=True,
        ),
        "kex_bc": ParamLocalSetting(
            name_setting=NameSetting("kex_bc", "", TPL),
            min=0.0,
            value=200.0,
            vary=True,
        ),
        "pa": ParamLocalSetting(
            name_setting=NameSetting("pa", "", TPL),
            expr="1.0 - {pb} - {pc}",
        ),
        **kij_settings,
    }


def make_settings_3st_fork(_conditions: Conditions) -> dict[str, ParamLocalSetting]:
    kij_settings = create_kij_settings("abc")
    del kij_settings["kbc"]
    del kij_settings["kcb"]
    return {
        "pb": ParamLocalSetting(
            name_setting=NameSetting("pb", "", TPL),
            value=0.02,
            min=0.0,
            max=1.0,
            vary=True,
        ),
        "pc": ParamLocalSetting(
            name_setting=NameSetting("pc", "", TPL),
            value=0.02,
            min=0.0,
            max=1.0,
            vary=True,
        ),
        "kex_ab": ParamLocalSetting(
            name_setting=NameSetting("kex_ab", "", TPL),
            min=0.0,
            value=200.0,
            vary=True,
        ),
        "kex_ac": ParamLocalSetting(
            name_setting=NameSetting("kex_ac", "", TPL),
            min=0.0,
            value=200.0,
            vary=True,
        ),
        "pa": ParamLocalSetting(
            name_setting=NameSetting("pa", "", TPL),
            expr="1.0 - {pb} - {pc}",
        ),
        **kij_settings,
    }


def make_settings_3st(conditions: Conditions) -> dict[str, ParamLocalSetting]:
    return {
        **make_settings_3st_fork(conditions),
        **make_settings_3st_linear(conditions),
    }


def make_settings_4st(conditions: Conditions) -> dict[str, ParamLocalSetting]:
    return {
        **make_settings_3st(conditions),
        "pd": ParamLocalSetting(
            name_setting=NameSetting("pd", "", TPL),
            value=0.02,
            min=0.0,
            max=1.0,
            vary=True,
        ),
        "kex_ad": ParamLocalSetting(
            name_setting=NameSetting("kex_ad", "", TPL),
            min=0.0,
            value=0.0,
        ),
        "kex_bd": ParamLocalSetting(
            name_setting=NameSetting("kex_bd", "", TPL),
            min=0.0,
            value=0.0,
        ),
        "kex_cd": ParamLocalSetting(
            name_setting=NameSetting("kex_cd", "", TPL),
            min=0.0,
            value=200.0,
            vary=True,
        ),
        "pa": ParamLocalSetting(
            name_setting=NameSetting("pa", "", TPL),
            min=0.0,
            max=1.0,
            expr="1.0 - {pb} - {pc} - {pd}",
        ),
        **create_kij_settings("abcd"),
    }


def register() -> None:
    model_factory.register(name="3st", setting_maker=make_settings_3st)
    model_factory.register(name="3st_triangle", setting_maker=make_settings_3st)
    model_factory.register(name="3st_linear", setting_maker=make_settings_3st_linear)
    model_factory.register(name="3st_fork", setting_maker=make_settings_3st_fork)
    model_factory.register(name="4st", setting_maker=make_settings_4st)
